/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: riskgame.c,v 1.47 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * "World" used by server as well as client
 */

static const char
    rcsid[] = "@(#)$Id: riskgame.c,v 1.47 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "network.h"
#include "riskgame.h"
#include "deck.h"
#include "clients.h"
#include "debug.h"
#include "language.h"

/* Shared variables */
GAME_States iGameState  = GAME_REGISTERING;

/* Private variables */
static FILE *hLogFile = ( FILE * ) NULL;

/* Private functions */
static Int32    _RISK_GetMsgType( Int32 iField );
static Int32    _RISK_GetIntValue( Int32 iField, Int32 iIndex1, Int32 iIndex2 );
static CString  _RISK_GetStrValue( Int32 iField, Int32 iIndex1, Int32 iIndex2 );

static Msgs    *_RISK_BuildMsg( Int32 iMessType,
                                Int32 iField,
                                Int32 iIndex1,
                                Int32 iIndex2,
                                CString strNewValue,
                                Int32 iNewValue );

static void     _RISK_Replicate( Int32 iField,
                                 Int32 iIndex1,
                                 Int32 iIndex2,
                                 CString strNewValue,
                                 Int32 iNewValue );

/**
 * Continent
 */
typedef struct
    {
    CString     strName;            /**< Name of the continent */
    Int32       iValue;             /**< Number of armies it's worth */
    Int32       iNumCountries;      /**< Number of countries in it */
    } ContinentObject;

/**
 * Country
 */
typedef struct
    {
    CString     strName;            /**< Name of the country */
    Int32       iContinent;         /**< The continent it forms part of */
    Int32       iNumArmies;         /**< Number of armies on the country */
    Int32       iTextX;             /**< X coordinate of country label */
    Int32       iTextY;             /**< Y coordinate of country label */
    Int32       piOwner;            /**< The player it belongs to */
    Int32       piAdjCountries[ 6 ];  /**< All attackable countries */
    } CountryObject;

/**
 * Player
 */
typedef struct
    {	
    Int32       iAllocationState;   /**< Allocation state */
    Int32       iAttackMode;        /**< Attack mode */
    Int32       iDiceMode;          /**< Dice roll mode */
    Int32       iMsgDstMode;        /**< Message destination */
    Int32       iSpecies;           /**< Species ( human/some AI) */
    Flag        iState;             /**< State */
    Int32       iClient;            /**< Client ID */
    CString     strName;            /**< Name */
    CString     strColor;           /**< Color string */
    Int32       iCountriesOwned;    /**< Number of countries owned */
    Int32       iNumArmies;         /**< Number of armies on the board */
    Int32       iNumCards;          /**< Number of cards in hand */
    Int32       piCards[ MAX_CARDS ]; /**< Card deck */
    Int16       typOfMission;       /**< Type of mission */
    Int32       mission1;           /**< Mission 1 */
    Int32       mission2;           /**< Mission 2 */
    } PlayerObject;

/**
 * Species
 */
typedef struct SpeciesObject
    {
    Int32       iAllocationState;   /**< Allocation state */
    Int32       iClient;            /**< Client ID */
    CString     strName;            /**< Name */
    CString     strAuthor;          /**< Author */
    CString     strVersion;         /**< Version */
    CString     strDescription;     /**< Description */
    struct SpeciesObject *next;     /**< Pointer to next */
    } SpeciesObject;



/**
 * Risk game object
 */
typedef struct
    {
    /** Replication callback pointer */
    void                ( *ReplicateCallback ) ( const Msgs *, Int32, Int32 );
    /** Failure callback pointer */
    void                ( *FailureCallback ) ( CString, Int32 );
    /** Pre-view callback pointer */
    void                ( *PreViewCallback ) ( const Msgs * );
    /** Post-view callback pointer */
    void                ( *PostViewCallback ) ( const Msgs * );

    /** Species list */
    SpeciesObject       pSpecies;
    /** Continents */
    ContinentObject     pContinents[ NUM_CONTINENTS ];
    /** Players */
    PlayerObject        pPlayers[ MAX_PLAYERS ];
    /** Countries */
    CountryObject       pCountries[ NUM_COUNTRIES ];
    } Game;

/**
 * Hardcoded Risk game contents
 *
 * \bug Map should be loaded from a file
 */
static Game RiskGame =
    {
    NULL, NULL, NULL, NULL,
    {
    ALLOC_COMPLETE,
    -1, HUMAN, UNKNOWN,
    "0.001",
    PINK_CREATURE,
    ( SpeciesObject * ) NULL
    },
    {
    { NORTH_AMERICA, 5, 9 },
    { SOUTH_AMERICA, 2, 4 },
    { AFRICA,        3, 6 },
    { AUSTRALIA,     2, 4 },
    { ASIA,          7, 12 },
    { EUROPE,        5, 7 },
    },
    {
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1  },
    { ALLOC_NONE, 1, 3, 0, 0, PLAYER_ALIVE, -1, NULL, NULL, 0, 0, 0,
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, NO_MISSION, -1, -1 },
    },
    {

    { GREENLAND,  CNT_NORTHAMERICA,     0, 264,  32, -1, /* 0 */
      { 1, 5, 10, 11, -1, -1 } },
    { ICELAND,  CNT_EUROPE,             0, 342,  51, -1,
      { 0, 7, 14, -1, -1, -1 } },
    { SIBERIA,  CNT_ASIA,               0, 622,  91, -1,
      { 3, 6, 13, 21, 22, -1 } },
    { URAL,  CNT_ASIA,                  0, 557, 106, -1,
      { 2, 8, 15, 22, -1, -1 } },
    { ALASKA,  CNT_NORTHAMERICA,        0,  13,  61, -1,
      { 5, 9, 12, -1, -1, -1 } },
    { NORTHWEST,  CNT_NORTHAMERICA,     0, 61, 62, -1, /* 5 */
      { 0, 4, 11, 12, -1, -1 } },
    { IRKUTSK,  CNT_ASIA,               0, 700,  71, -1,
      { 2, 9, 13, -1, -1, -1 } },
    { SCANDINAVIA, CNT_EUROPE,          0, 387,  85, -1,
      { 1, 8, 14, 16, -1, -1 } },
    { UKRAINE, CNT_EUROPE,              0, 450, 103, -1,
      { 3, 7, 15, 16, 20, 24 }},
    { KAMCHATKA, CNT_ASIA,              0, 731, 121, -1,
      { 4, 6, 13, 21, 23, -1 } },
    { QUEBEC, CNT_NORTHAMERICA,         0, 181,  95, -1,     /* 10 */
      { 0, 11, 18, -1, -1, -1 } },
    { ONTARIO, CNT_NORTHAMERICA,        0, 116,  97, -1,
      { 0, 5, 10, 12, 17, 18 } },
    { ALBERTA, CNT_NORTHAMERICA,        0,  55,  90, -1,
      { 4, 5, 11, 17, -1, -1 } },
    { YAKUTSK, CNT_ASIA,                0, 690, 111, -1,
      { 2, 6, 9, 21, -1, -1 } },
    { GREAT_BRITAIN, CNT_EUROPE,        0, 338, 110, -1,
      { 1, 7, 16, 19, -1, -1 } },
    { AFGHANISTAN, CNT_ASIA,            0, 534, 141, -1,  /* 15 */
      { 3, 8, 22, 24, 28, -1 } },
    { NORTHERN_EUROPE, CNT_EUROPE,      0, 386, 113, -1,
      { 7, 8, 14, 19, 20, -1 } },
    { WEST_STATES, CNT_NORTHAMERICA,    0, 62, 126, -1,
      { 11, 12, 18, 25, -1, -1 } },
    { EAST_STATES, CNT_NORTHAMERICA,    0, 126, 145, -1,
      { 10, 11, 17, 25, -1, -1 } },
    { WEST_EUROPE, CNT_EUROPE,          0, 338, 148, -1,
      { 14, 16, 20, 26, -1, -1 } },
    { SOUTH_EUROPE, CNT_EUROPE,         0, 416, 134, -1, /* 20 */
      { 8, 16, 19, 24, 26, 27 } },
    { MONGOLIA, CNT_ASIA,               0, 685, 139, -1,
      { 2, 9, 13, 22, 23, -1 } },
    { CHINA,  CNT_ASIA,                 0, 632, 163, -1,
      { 2, 3, 15, 21, 28, 29 } },
    { JAPAN, CNT_ASIA,                  0, 735, 166, -1,
      { 9, 21, -1, -1, -1, -1 } },
    { MIDDLE_EAST, CNT_ASIA,            0, 477, 193, -1,
      { 8, 15, 20, 27, 28, 30 } },
    { CENTRAL_AMERICA, CNT_NORTHAMERICA,0, 87, 165, -1, /* 25 */
      { 17, 18, 32, -1, -1, -1 }},
    { NORTH_AFRICA, CNT_AFRICA,         0, 339, 214, -1,
      { 19, 20, 27, 30, 33, 35 } },
    { EGYPT, CNT_AFRICA,                0, 407, 194, -1,
      { 20, 24, 26, 30, -1, -1 } },
    { INDIA, CNT_ASIA,                  0, 572, 194, -1,
      { 15, 22, 24, 29, -1, -1 } },
    { SIAM,  CNT_ASIA,                  0, 637, 207, -1,
      { 22, 28, 31, -1, -1, -1 } },
    { EAST_AFRICA, CNT_AFRICA,          0, 437, 231, -1, /* 30 */
      { 24, 26, 27, 35, 37, 40 } },
    { INDONESIA, CNT_AUSTRALIA,         0, 672, 262, -1,
      { 29, 36, 38, -1, -1, -1 } },
    { VENEZUELA, CNT_SOUTHAMERICA,      0, 188, 230, -1,
      { 25, 33, 34, -1, -1, -1 } },
    { BRASIL, CNT_SOUTHAMERICA,         0, 233, 277, -1,
      { 26, 32, 34, 41, -1, -1 } },
    { PERU, CNT_SOUTHAMERICA,           0, 194, 294, -1,
      { 32, 33, 41, -1, -1, -1 }},
    { CONGO, CNT_AFRICA,                0, 410, 271, -1, /* 35 */
      { 26, 30, 37, -1, -1, -1 } },
    { NEW_GUINEA, CNT_AUSTRALIA,        0, 751, 276, -1,
      { 31, 38, 39, -1, -1, -1 } },
    { SOUTH_AFRICA, CNT_AFRICA,         0, 416, 326, -1,
      { 30, 35, 40, -1, -1, -1 } },
    { WESTERN_AUSTRALIA, CNT_AUSTRALIA, 0, 700, 335, -1,
      { 31, 36, 39, -1, -1, -1 } },
    { EASTERN_AUSTRALIA, CNT_AUSTRALIA, 0, 756, 343, -1,
      { 36, 38, -1, -1, -1, -1 } },
    { MADAGASCAR, CNT_AFRICA,           0, 479, 321, -1, /* 40 */
      { 30, 37, -1, -1, -1, -1 } },
    { ARGENTINA,  CNT_SOUTHAMERICA,     0, 189, 352, -1,
      { 33, 34, -1, -1, -1, -1 } },
    }
    };

/* Species related private function */
SpeciesObject *_RISK_GetSpecies( Int32 iHandle );


/**
 * Init callbacks
 */
void RISK_InitObject( void ( *ReplicateCallback )( const Msgs *, Int32, Int32 ),
                      void ( *PreViewCallback )( const Msgs * ),
                      void ( *PostViewCallback )( const Msgs * ),
                      void ( *FailureCallback )( CString, Int32 ),
                      FILE *hDebug )
    {

    RiskGame.ReplicateCallback = ReplicateCallback;
    RiskGame.FailureCallback   = FailureCallback;
    RiskGame.PreViewCallback   = PreViewCallback;
    RiskGame.PostViewCallback  = PostViewCallback;

    /* The message callback has to be set -- it handles replication. */
    assert( RiskGame.ReplicateCallback );

    /* Save this */
    hLogFile = hDebug;
    }


/**
 * _RISK_Replicate
 */
void _RISK_Replicate( Int32 iField, Int32 iIndex1, Int32 iIndex2,
                      CString strNewValue, Int32 iNewValue )
    {
    Msgs   *pMess;
    Int32   iMessType;

    /* Get the message and the message type */
    iMessType   = _RISK_GetMsgType( iField );
    pMess       = _RISK_BuildMsg( iMessType,
                                  iField,
                                  iIndex1,
                                  iIndex2,
                                  strNewValue,
                                  iNewValue );
    assert( pMess );

    /* Call the callback for messages ( before changing dist. obj. ) */
    if ( RiskGame.PreViewCallback )
        {
        RiskGame.PreViewCallback( pMess );
        }

    /* Actually set the value ( sort of roundabout : ) Everything
     * to set the message follows the same code path now, so
     * it may be a bit roundabout but at least it's consistant.
     */

    RISK_ProcessMessage( pMess );

    /* Call the callback for messages ( after changing dist. obj. ) */
    if ( RiskGame.PostViewCallback )
        {
        RiskGame.PostViewCallback( pMess );
        }

    /* Call the message handler.  If there wasn't one, then
     * we would have a problem, since it handles the actual replication.
     * This involves sending a message to the server, or broadcasting
     * a message out, depending on whether a client or server has
     * registered this callback.  The last parameter doesn't matter in
     * this case, because it's an outgoing message ( the last parameter
     * is the message source ).
     */

    RiskGame.ReplicateCallback( pMess, MESS_OUTGOING, -1 );

    /* Delete the memory the message was taking */
    MEM_Free( pMess );
    }


/**
 * RISK_SelectiveReplicate
 */
void RISK_SelectiveReplicate( Int32 iSocket,
                              Int32 iField,
                              Int32 iIndex1,
                              Int32 iIndex2 )
    {
    Msgs   *pMess;
    Int32   iMessType;
    Int32   iValue = 0;
    CString strValue = NULL;

    /* Type of the message */
    iMessType = _RISK_GetMsgType( iField );

    /* Get the value... */
    if ( iMessType == MSG_OBJINTUPDATE )
        {
        iValue = _RISK_GetIntValue( iField, iIndex1, iIndex2 );
        }
    else
        {
        strValue = _RISK_GetStrValue( iField, iIndex1, iIndex2 );
        }

    /* And build the message... */
    pMess = _RISK_BuildMsg( iMessType,
                            iField,
                            iIndex1,
                            iIndex2,
                            strValue,
                            iValue );

    RISK_SendMessage( iSocket, pMess );

    /* Delete the memory the message took up */
    MEM_Free( pMess );
    }


/**
 * RISK_ResetObj
 */
void RISK_ResetObj( void )
    {
    Int32   i;
    Int32   iPlayer;

    /* Init the needed fields of the players */
    for ( i = 0; i < RISK_GetNumPlayers( ); i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );

        RISK_SetAttackModeOfPlayer( iPlayer, 1 );
        RISK_SetStateOfPlayer( iPlayer, PLAYER_ALIVE );
        RISK_SetClientOfPlayer( iPlayer, -1 );
        RISK_SetNumCountriesOfPlayer( iPlayer, 0 );
        RISK_SetNumArmiesOfPlayer( iPlayer, 0 );
        RISK_SetNumCardsOfPlayer( iPlayer, 0 );
        RISK_SetAllocationStateOfPlayer( iPlayer, ALLOC_NONE );
        RISK_SetMissionTypeOfPlayer( iPlayer, NO_MISSION );
        }

    /* Init the needed fields of the countries */
    for ( i = 0; i != NUM_COUNTRIES; i++ )
        {
        RISK_SetOwnerOfCountry( i, -1 );
        RISK_SetNumArmiesOfCountry( i, 0 );
        }
    }


/**
 * RISK_ResetGame
 */
void RISK_ResetGame( void )
    {
    Int32   i;
    Int32   iPlayer;

    /* Init the needed fields of the players */
    for ( i = 0; i < RISK_GetNumPlayers( ); i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );

        RISK_SetStateOfPlayer( iPlayer, PLAYER_ALIVE );
        RISK_SetNumCountriesOfPlayer( iPlayer, 0 );
        RISK_SetNumCardsOfPlayer( iPlayer, 0 );
        RISK_SetNumArmiesOfPlayer( iPlayer, 0 );
        RISK_SetMissionTypeOfPlayer( iPlayer, NO_MISSION );
        }

    /* Init the needed fields of the countries */
    for ( i = 0; i != NUM_COUNTRIES; i++ )
        {
        RISK_SetOwnerOfCountry( i, -1 );
        RISK_SetNumArmiesOfCountry( i, 0 );
        }
    }


/**
 * RISK_ProcessMessage
 */
void RISK_ProcessMessage( const Msgs *pMess )
    {
    Int32                   iIndex1;
    Int32                   iIndex2;
    Int32                   iValue;
    Int32                   iField;
    CString                 strValue;
    const MsgObjIntUpdate  *pIntMess = &( pMess->msgObjIntUpdate );
    const MsgObjStrUpdate  *pStrMess = &( pMess->msgObjStrUpdate );
    SpeciesObject          *s;

    assert( pMess );

    if ( pMess->msgId == MSG_OBJINTUPDATE )
        {
        iField  = pIntMess->iField;
        iIndex1 = pIntMess->iIndex1;
        iIndex2 = pIntMess->iIndex2;
        iValue  = pIntMess->iNewValue;

        switch ( iField )
            {
            case PLR_ATTACKMODE:
                RiskGame.pPlayers[ iIndex1 ].iAttackMode = iValue;
                break;

            case PLR_DICEMODE:
                RiskGame.pPlayers[ iIndex1 ].iDiceMode = iValue;
                break;

            case PLR_MSGDSTMODE:
                RiskGame.pPlayers[ iIndex1 ].iMsgDstMode = iValue;
                break;

            case PLR_STATE:
                RiskGame.pPlayers[ iIndex1 ].iState = iValue;
                break;

            case PLR_CLIENT:
                RiskGame.pPlayers[ iIndex1 ].iClient = iValue;
                break;

            case PLR_NUMCOUNTRIES:
                RiskGame.pPlayers[ iIndex1 ].iCountriesOwned = iValue;
                break;

            case PLR_NUMARMIES:
                RiskGame.pPlayers[ iIndex1 ].iNumArmies = iValue;
                break;

            case PLR_NUMCARDS:
                RiskGame.pPlayers[ iIndex1 ].iNumCards = iValue;
                break;

            case PLR_SPECIES:
                RiskGame.pPlayers[ iIndex1 ].iSpecies = iValue;
                break;

            case PLR_ALLOCATION:
                RiskGame.pPlayers[ iIndex1 ].iAllocationState = iValue;
                break;

            case PLR_CARD:
                /* TdH: one of the places where greenlandbug might be caught */
                RiskGame.pPlayers[ iIndex1 ].piCards[ iIndex2 ] = iValue;
                break;

            case PLR_MISSION:
                RiskGame.pPlayers[ iIndex1 ].typOfMission = iValue;
                break;

            case PLR_MISSION1:
                RiskGame.pPlayers[ iIndex1 ].mission1 = iValue;
                break;

            case PLR_MISSION2:
                RiskGame.pPlayers[ iIndex1 ].mission2 = iValue;
                break;

            case CNT_NUMARMIES:
                RiskGame.pCountries[ iIndex1 ].iNumArmies = iValue;
                break;

            case CNT_OWNER:
                RiskGame.pCountries[ iIndex1 ].piOwner = iValue;
                break;

            case SPE_CLIENT:
                s = _RISK_GetSpecies( iIndex1 );
                s->iClient = iValue;
                break;

            case SPE_ALLOCATION:
                s = _RISK_GetSpecies( iIndex1 );
                s->iAllocationState = iValue;
                break;
	
            default:
                assert( FALSE );
                break;
            }
        }
    else
        {
        iField   = pStrMess->iField;
        iIndex1  = pStrMess->iIndex1;
        iIndex2  = pStrMess->iIndex2;
        strValue = pStrMess->strNewValue;

        switch ( iField )
            {
            case PLR_NAME:
                if ( RiskGame.pPlayers[ iIndex1 ].strName != NULL )
                    {
                    MEM_Free( RiskGame.pPlayers[ iIndex1 ].strName );
                    }

                if ( strValue )
                    {
                    RiskGame.pPlayers[ iIndex1 ].strName =
                        ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                    strcpy( RiskGame.pPlayers[ iIndex1 ].strName, strValue );
                    }
                else
                    {
                    RiskGame.pPlayers[ iIndex1 ].strName = NULL;
                    }
                break;
            case PLR_COLORSTRING:
                if ( RiskGame.pPlayers[ iIndex1 ].strColor != NULL )
                    {
                    MEM_Free( RiskGame.pPlayers[ iIndex1 ].strColor );
                    }

                if ( strValue )
                    {
                    RiskGame.pPlayers[ iIndex1 ].strColor =
                        ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                    strcpy( RiskGame.pPlayers[ iIndex1 ].strColor, strValue );
                    }
                else
                    {
                    RiskGame.pPlayers[ iIndex1 ].strColor = NULL;
                    }
                break;
            case SPE_NAME:
            {
            SpeciesObject *s = _RISK_GetSpecies( iIndex1 );
	
            if ( s->strName != NULL )
                {
                MEM_Free( s->strName );
                }
	
            if ( strValue )
                {
                s->strName = ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                strcpy( s->strName, strValue );
                }
            else
                {
                s->strName = NULL;
                }
            }
            break;
            case SPE_VERSION:
            {
            SpeciesObject *s = _RISK_GetSpecies( iIndex1 );
	
            if ( s->strVersion != NULL )
                {
                MEM_Free( s->strVersion );
                }
	
            if ( strValue )
                {
                s->strVersion = ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                strcpy( s->strVersion, strValue );
                }
            else
                {
                s->strVersion = NULL;
                }
            }
            break;
            case SPE_DESCRIPTION:
            {
            SpeciesObject *s = _RISK_GetSpecies( iIndex1 );
	
            if ( s->strDescription != NULL )
                {
                MEM_Free( s->strDescription );
                }
	
            if ( strValue )
                {
                s->strDescription = ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                strcpy( s->strDescription, strValue );
                }
            else
                {
                s->strDescription = NULL;
                }
            }
            break;
            case SPE_AUTHOR:
            {
            SpeciesObject *s = _RISK_GetSpecies( iIndex1 );
	
            if ( s->strAuthor != NULL )
                {
                MEM_Free( s->strAuthor );
                }
	
            if ( strValue )
                {
                s->strAuthor = ( CString ) MEM_Alloc( strlen( strValue ) + 1 );
                strcpy( s->strAuthor, strValue );
                }
            else
                {
                s->strAuthor = strValue;
                }
            }
            break;

            default:
                assert( FALSE );
                break;
            }
        }
    }

/**
 * RISK_SetAttackModeOfPlayer
 */
void RISK_SetAttackModeOfPlayer( Int32 iPlayer, Int32 iMode )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_ATTACKMODE, iPlayer, 0, NULL, iMode );
    }

/**
 * RISK_SetDIceModeOfPlayer
 */
void RISK_SetDiceModeOfPlayer( Int32 iPlayer, Int32 iMode )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_DICEMODE, iPlayer, 0, NULL, iMode );
    }

/**
 * RISK_SetMsgDestModeOfPlayer
 */
void RISK_SetMsgDstModeOfPlayer( Int32 iPlayer, Int32 iMode )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MSGDSTMODE, iPlayer, 0, NULL, iMode );
    }

/**
 * RISK_SetStateOfPlayer
 */
void RISK_SetStateOfPlayer( Int32 iPlayer, Int32 iState )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_STATE, iPlayer, 0, NULL, iState );
    }

/**
 * RISK_SetClientOfPlayer
 */
void RISK_SetClientOfPlayer( Int32 iPlayer, Int32 iClient )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_CLIENT, iPlayer, 0, NULL, iClient );
    }

/**
 * RISK_SetNumCountriesOfPlayer
 */
void RISK_SetNumCountriesOfPlayer( Int32 iPlayer, Int32 iNumCountries )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_NUMCOUNTRIES, iPlayer, 0, NULL, iNumCountries );
    }

/**
 * RISK_SetNumArmiesOfPlayer
 */
void RISK_SetNumArmiesOfPlayer( Int32 iPlayer, Int32 iNumArmies )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_NUMARMIES, iPlayer, 0, NULL, iNumArmies );
    }

/**
 * RISK_SetNumCardsOfPlayer
 */
void RISK_SetNumCardsOfPlayer( Int32 iPlayer, Int32 iNumCards )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_NUMCARDS, iPlayer, 0, NULL, iNumCards );
    }

/**
 * RISK_SetNameOfPlayer
 */
void RISK_SetNameOfPlayer( Int32 iPlayer, CString strName )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_NAME, iPlayer, 0, strName, 0 );
    }

/**
 * RISK_SetColorCStringOfPlayer
 */
void RISK_SetColorCStringOfPlayer( Int32 iPlayer, CString strColor )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_COLORSTRING, iPlayer, 0, strColor, 0 );
    }

/**
 * RISK_SetCardOfPlayer
 */
void RISK_SetCardOfPlayer( Int32 iPlayer, Int32 iCard, Int32 iValue )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_CARD, iPlayer, iCard, NULL, iValue );
    }

/**
 * RISK_SetMissionTypeOfPlayer
 */
void RISK_SetMissionTypeOfPlayer( Int32 iPlayer, Int16 typ )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION, iPlayer, 0, NULL, typ );
    }

/**
 * RISK_SetMissionNumberOfPlayer
 */
void RISK_SetMissionNumberOfPlayer( Int32 iPlayer, Int32 n )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION1, iPlayer, 0, NULL, n );
    }

/**
 * RISK_SetMissionContinent1OfPlayer
 */
void RISK_SetMissionContinent1OfPlayer( Int32 iPlayer, Int32 n )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION1, iPlayer, 0, NULL, n );
    }

/**
 * RISK_SetMissionContinent2OfPlayer
 */
void RISK_SetMissionContinent2OfPlayer( Int32 iPlayer, Int32 n )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION2, iPlayer, 0, NULL, n );
    }

/**
 * RISK_SetMissionMissionPlayerToKillOfPlayer
 */
void RISK_SetMissionMissionPlayerToKillOfPlayer( Int32 iPlayer, Int32 n )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION1, iPlayer, 0, NULL, n );
    }

/**
 * RISK_SetMissionPlayerIsKilledOfPlayer
 */
void RISK_SetMissionPlayerIsKilledOfPlayer( Int32 iPlayer, Flag boool )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_MISSION2, iPlayer, 0, NULL, boool );
    }

/**
 * RISK_SetSpeciesOfPlayer
 */
void RISK_SetSpeciesOfPlayer( Int32 iPlayer, Int32 iSpecies )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_SPECIES, iPlayer, 0, NULL, iSpecies );
    }

/**
 * RISK_SetAllocationStateOfPlayer
 */
void RISK_SetAllocationStateOfPlayer( Int32 iPlayer, Int32 iState )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    _RISK_Replicate( PLR_ALLOCATION, iPlayer, 0, NULL, iState );
    }

/**
 * RISK_SetNumArmiesOfCountry
 */
void RISK_SetNumArmiesOfCountry( Int32 iCountry, Int32 iNumArmies )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );

    _RISK_Replicate( CNT_NUMARMIES, iCountry, 0, NULL, iNumArmies );
    }

/**
 * RISK_SetOwnerOfCountry
 */
void RISK_SetOwnerOfCountry( Int32 iCountry, Int32 iOwner )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );

    _RISK_Replicate( CNT_OWNER, iCountry, 0, NULL, iOwner );
    }

/**
 * RISK_SetClientOfSpecies
 */
void RISK_SetClientOfSpecies( Int32 iHandle, Int32 iClient )
    {
    _RISK_Replicate( SPE_CLIENT, iHandle, 0, NULL, iClient );
    }

/**
 * RISK_SetNameOfSpecies
 */
void RISK_SetNameOfSpecies( Int32 iHandle, CString strName )
    {
    _RISK_Replicate( SPE_NAME, iHandle, 0, strName, 0 );
    }

/**
 * RISK_SetAllocationStateOfSpecies
 */
void RISK_SetAllocationStateOfSpecies( Int32 iHandle, Int32 iState )
    {
    _RISK_Replicate( SPE_ALLOCATION, iHandle, 0, NULL, iState );
    }

/**
 * RISK_SetAuthorOfSpecies
 */
void RISK_SetAuthorOfSpecies( Int32 iHandle, CString strAuthor )
    {
    _RISK_Replicate( SPE_AUTHOR, iHandle, 0, strAuthor, 0 );
    }

/**
 * RISK_SetVersionOfSpecies
 */
void RISK_SetVersionOfSpecies( Int32 iHandle, CString strVersion )
    {
    _RISK_Replicate( SPE_VERSION, iHandle, 0, strVersion, 0 );
    }

/**
 * RISK_SetDescriptionOfSpecies
 */
void RISK_SetDescriptionOfSpecies( Int32 iHandle, CString strDescription )
    {
    _RISK_Replicate( SPE_DESCRIPTION, iHandle, 0, strDescription, 0 );
    }

/**
 * Determine number of initialized players
 */
Int32 RISK_GetNumPlayers( void )
    {
    Int32   i;
    Int32   iCount;

    for ( i = iCount = 0; i < MAX_PLAYERS; i++ )
        {
        if ( RISK_GetAllocationStateOfPlayer( i ) == ALLOC_COMPLETE )
            {
            iCount++;
            }
        }

    return iCount;
    }

/**
 * RISK_GetAttackModeOfPlayer
 */
Int32 RISK_GetAttackModeOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iAttackMode;
    }

/**
 * RISK_GetMsgDstModeOfPlayer
 */
Int32 RISK_GetMsgDstModeOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iMsgDstMode;
    }

/**
 * RISK_GetDiceModeOfPlayer
 */
Int32 RISK_GetDiceModeOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iDiceMode;
    }

/**
 * RISK_GetStateOfPlayer
 */
Int32 RISK_GetStateOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iState;
    }

/**
 * RISK_GetClientOfPlayer
 */
Int32 RISK_GetClientOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer < MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iClient;
    }

/**
 * RISK_GetAlocationStateOfPlayer
 */
Int32 RISK_GetAllocationStateOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iAllocationState;
    }

/**
 * RISK_GetNumCountriesOfPlayer
 */
Int32 RISK_GetNumCountriesOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iCountriesOwned;
    }

/**
 * RISK_GetNumArmiesOfPlayer
 */
Int32 RISK_GetNumArmiesOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iNumArmies;
    }

/**
 * RISK_GetNumCardsOfPlayer
 */
Int32 RISK_GetNumCardsOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iNumCards;
    }

/**
 *
 */
Int32 RISK_GetSpeciesOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].iSpecies;
    }

/**
 *
 */
Int32 RISK_GetValueOfContinent( Int32 iContinent )
    {
    assert( iContinent >= 0 && iContinent< NUM_CONTINENTS );

    return RiskGame.pContinents[ iContinent ].iValue;
    }

/**
 *
 */
CString RISK_GetNameOfContinent( Int32 iContinent )
    {
    assert( iContinent >= 0 && iContinent< NUM_CONTINENTS );

    return RiskGame.pContinents[ iContinent ].strName;
    }

/**
 *
 */
Int32 RISK_GetNumCountriesOfContinent( Int32 iContinent )
    {
    assert( iContinent >= 0 && iContinent< NUM_CONTINENTS );

    return RiskGame.pContinents[ iContinent ].iNumCountries;
    }

/**
 *
 */
CString RISK_GetNameOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    /*
     * This patches the symptom of a problem whose cause I haven't found yet.
     * -- Pac.
     *
     *  return RiskGame.pPlayers[ iPlayer ].strName;
     */
    if ( RiskGame.pPlayers[ iPlayer ].strName )
        {
        return ( RiskGame.pPlayers[ iPlayer ].strName );
        }
    else
        {
        return ( "UnknownPlayerBug" );
        }
    }

/**
 *
 */
CString RISK_GetColorCStringOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].strColor;
    }

/**
 *
 */
Int32 RISK_GetCardOfPlayer( Int32 iPlayer, Int32 iCard )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].piCards[ iCard ];
    }

/**
 *
 */
Int16 RISK_GetMissionTypeOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].typOfMission;
    }

/**
 *
 */
Int32 RISK_GetMissionContinent1OfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].mission1;
    }

/**
 *
 */
Int32 RISK_GetMissionNumberOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].mission1;
    }

/**
 *
 */
Int32 RISK_GetMissionContinent2OfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].mission2;
    }

/**
 *
 */
Int32 RISK_GetMissionPlayerToKillOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].mission1;
    }

/**
 *
 */
Flag RISK_GetMissionIsPlayerKilledOfPlayer( Int32 iPlayer )
    {
    assert( iPlayer >= 0 && iPlayer< MAX_PLAYERS );

    return RiskGame.pPlayers[ iPlayer ].mission2;
    }

/**
 *
 */
CString RISK_GetNameOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry< NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].strName;
    }

/**
 *
 */
Int32 RISK_GetContinentOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry< NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].iContinent;
    }

/**
 *
 */
Int32 RISK_GetNumArmiesOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry< NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].iNumArmies;
    }

/**
 *
 */
Int32 RISK_GetOwnerOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry< NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].piOwner;
    }

/**
 *
 */
Int32 RISK_GetAdjCountryOfCountry( Int32 iCountry, Int32 iIndex )
    {
    assert( iCountry >= 0 && iCountry< NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].piAdjCountries[ iIndex ];
    }

/**
 *
 */
Int32 RISK_GetNumLivePlayers( void )
    {
    Int32   i;
    Int32   iCount;

    for ( i = iCount = 0; i < MAX_PLAYERS; i++ )
        {
        if ( RISK_GetAllocationStateOfPlayer( i ) == ALLOC_COMPLETE &&
             RISK_GetStateOfPlayer( i ) == PLAYER_ALIVE )
            {
            iCount++;
            }
        }

    return iCount;
    }

/**
 *
 */
Int32 RISK_GetNumSpecies( void )
    {
    Int32           iCount;
    SpeciesObject  *pSpecies;

    for ( pSpecies = &RiskGame.pSpecies, iCount = 0;
          pSpecies != NULL;
          pSpecies = pSpecies->next )
        {
        if ( pSpecies->iAllocationState == ALLOC_COMPLETE )
            {
            iCount++;
            }
        }

    return iCount;
    }

/**
 *
 */
Int32 RISK_GetTextXOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].iTextX;
    }

/**
 *
 */
Int32 RISK_GetTextYOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );

    return RiskGame.pCountries[ iCountry ].iTextY;
    }

/**
 *
 */
Int32 RISK_GetClientOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->iClient;
    }

/**
 *
 */
CString RISK_GetNameOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->strName;
    }

/**
 *
 */
Int32 RISK_GetAllocationStateOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->iAllocationState;
    }

/**
 *
 */
CString RISK_GetAuthorOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->strAuthor;
    }

/**
 *
 */
CString RISK_GetDescriptionOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->strDescription;
    }

/**
 *
 */
CString RISK_GetVersionOfSpecies( Int32 iHandle )
    {
    return _RISK_GetSpecies( iHandle )->strVersion;
    }

/**
 *
 */
Int32 _RISK_GetMsgType( Int32 iField )
    {
    Int32 retval = 0;

    switch ( iField )
        {
        case PLR_ATTACKMODE:
        case PLR_DICEMODE:
        case PLR_MSGDSTMODE:
        case PLR_STATE:
        case PLR_SPECIES:
        case PLR_ALLOCATION:
        case PLR_CLIENT:
        case PLR_NUMCOUNTRIES:
        case PLR_NUMARMIES:
        case PLR_NUMCARDS:
        case PLR_CARD:
        case PLR_MISSION:
        case PLR_MISSION1:
        case PLR_MISSION2:
        case CNT_CONTINENT:
        case CNT_NUMARMIES:
        case CNT_OWNER:
        case CNT_ADJCOUNTRY:
        case CNT_TEXTX:
        case CNT_TEXTY:
        case CON_VALUE:
        case CON_NUMCOUNTRIES:
        case SPE_CLIENT:
        case SPE_ALLOCATION:
            retval = MSG_OBJINTUPDATE;
            break;

        case PLR_NAME:
        case PLR_COLORSTRING:
        case CNT_NAME:
        case CON_NAME:
        case SPE_NAME:
        case SPE_VERSION:
        case SPE_DESCRIPTION:
        case SPE_AUTHOR:
            retval = MSG_OBJSTRUPDATE;
            break;

        default:
            assert( FALSE );
        }

    return retval;
    }

/**
 *
 */
Msgs *_RISK_BuildMsg( Int32 iMessType,
                      Int32 iField,
                      Int32 iIndex1,
                      Int32 iIndex2,
                      CString strValue,
                      Int32 iValue )
    {

    Msgs *pMess = NULL;
#ifndef NDEBUG
    printf( "RISK_BuildMsg called with iMessType %i\n", iMessType );
#endif /* NDEBUG */

    if ( iMessType == MSG_OBJINTUPDATE )
        {
        pMess = ( Msgs * ) MEM_Alloc( sizeof( *pMess ) );

        /* Build the new message */
        pMess->msgId                        = MSG_OBJINTUPDATE;
        pMess->msgObjIntUpdate.iField       = iField;
        pMess->msgObjIntUpdate.iIndex1      = iIndex1;
        pMess->msgObjIntUpdate.iIndex2      = iIndex2;
        pMess->msgObjIntUpdate.iNewValue    = iValue;
        }
    else if ( iMessType == MSG_OBJSTRUPDATE )
        {
        pMess = ( Msgs * ) MEM_Alloc( sizeof( *pMess ) );

        /* Build the new message */
        pMess->msgId                        = MSG_OBJSTRUPDATE;
        pMess->msgObjStrUpdate.iField       = iField;
        pMess->msgObjStrUpdate.iIndex1      = iIndex1;
        pMess->msgObjStrUpdate.iIndex2      = iIndex2;
        pMess->msgObjStrUpdate.strNewValue  = strValue;
        }
    else
        {
            /* Do nothing */
            assert( pMess == NULL );
        }



    return pMess;
    }


/**
 * RISK_ObjectFailure
 */
void RISK_ObjectFailure( CString strReason, Int32 iCommLink )
    {
    assert( RiskGame.FailureCallback );

    RiskGame.FailureCallback( strReason, iCommLink );
    }


/**
 * RISK_GetNthLivePlayer
 */
Int32 RISK_GetNthLivePlayer( Int32 iIndex )
    {
    Int32   i;
    Int32   iCount;
    Int32   retval = -1;

    assert( iIndex >= 0 && iIndex < RISK_GetNumLivePlayers( ) );

    /* Increment, because the index passed in will start at zero */
    iIndex++;

    /* Loop through, looking for the iIndex'th allocated player. */
    for ( i = 0, iCount = 0; i != MAX_PLAYERS && iIndex != iCount; i++ )
        {
        if ( ( RISK_GetAllocationStateOfPlayer( i ) == ALLOC_COMPLETE ) &&
             ( RISK_GetStateOfPlayer( i ) == PLAYER_ALIVE ) )
            {
            iCount++;
            }
        }

    if ( iIndex == iCount )
        {
        retval = i - 1;
        }

    /* Something wierd... */
    assert( retval != -1 );

    return retval;
    }


/**
 * RISK_GetNthPlayerAtClient
 */
Int32 RISK_GetNthPlayerAtClient( Int32 iClient, Int32 iIndex )
    {
    Int32   i;
    Int32   iCount;
    Int32   retval = -1;

    /* ok, seems to be harmless, guess this assertion is not needed, and even wrong
       printf( "RISK_GetNthPlayerAtClient got %d, numplayers returned %d. to fix!!\n", iIndex, RISK_GetNumPlayers( ));
    */
    /* tony: attention! ai's cause this one to fire */
    /*  printf( "before assertion\n" );
        assert( iIndex >= 0 && iIndex < RISK_GetNumPlayers( ) );
        printf( "passed assertion\n" );*/
    /* Increment, because the index passed in will start at zero */
    iIndex++;

    /* Loop through, looking for the iIndex'th player */
    for ( i = 0, iCount = 0; i != MAX_PLAYERS && iIndex != iCount; i++ )
        {
        if ( RISK_GetClientOfPlayer( i ) == iClient )
            {
            iCount++;
            }
        }

    if ( iIndex == iCount )
        {
        retval = i - 1;
        }

    assert( retval != -1 );

    return retval;
    }


/**
 * RISK_GetNthPlayer
 */
Int32 RISK_GetNthPlayer( Int32 iIndex )
    {
    Int32   i;
    Int32   iCount;
    Int32   retval = -2;

    assert( iIndex >= 0 && iIndex < RISK_GetNumPlayers( ) );

    /* Increment, because the index passed in will start at zero */
    iIndex++;

    /* Loop through, looking for the iIndex'th player */
    for ( i = 0, iCount = 0; i != MAX_PLAYERS && iIndex != iCount; i++ )
        {
        if ( RISK_GetAllocationStateOfPlayer( i ) == ALLOC_COMPLETE )
            {
            iCount++;
            }
        }

    if ( iIndex == iCount )
        {
        retval = i - 1;
        }

    /* Something wierd... */
    assert( retval != -1 );

    return retval;
    }


/**
 * RISK_SendMessage
 */
Int32 RISK_SendMessage( Int32 iDest, const Msgs *pMessage )
    {
    Int32   retval = 1;

    assert( pMessage != NULL );

    /* Call the callback for messages if it's not a dist. obj. related msg. */
    if ( ( pMessage->msgId != MSG_OBJINTUPDATE ) &&
         ( pMessage->msgId != MSG_OBJSTRUPDATE ) )
        {
        if ( RiskGame.PreViewCallback )
            {
            RiskGame.PreViewCallback( pMessage );
            }

        if ( RiskGame.PostViewCallback )
            {
            RiskGame.PostViewCallback( pMessage );
            }
        }

    /* Try to send it, if it fails, call the appropriate recovery routines */
    if ( NET_SendMessage( iDest, pMessage ) != 0 )
        {
        RISK_ObjectFailure( TXT_SEND_FAILED, iDest );
        retval = 0;
        }

    /* Log it if there's a file */
    if ( hLogFile )
        {
        fprintf( hLogFile,
                 "** Sent Message to ( %d ): %s\n",
                 iDest,
                 NET_MessageToString( pMessage ) );
        fflush( hLogFile );
        }

    return retval;
    }


/**
 * RISK_ReceiveMessage
 */
Int32 RISK_ReceiveMessage( Int32 iSource, Msgs **ppMessage )
    {
    /* Init these */
    *ppMessage = NULL;

    /* Try to receive it, if it fails, call the appropriate recovery routines */
    if ( NET_RecvMessage( iSource, ppMessage ) != 0 )
        {
        return 0;
        }

    /* Log it if there's a file */
    if ( hLogFile )
        {
        fprintf( hLogFile,
                 "** Rec. Message from ( %d ): %s\n",
                 iSource,
                 NET_MessageToString( *ppMessage ) );
        fflush( hLogFile );
        }

    /* Call the callback for messages ( before changing dist. obj. ) */
    if ( RiskGame.PreViewCallback )
        {
        RiskGame.PreViewCallback( *ppMessage );
        }

    /* If the message coming in is a dist. obj. msg., then don't
     * let it get through to the upper layers, as they shouldn't
     * need to get up there for any reason.
     */

    if ( ( ( *ppMessage )->msgId == MSG_OBJINTUPDATE ) ||
         ( ( *ppMessage )->msgId == MSG_OBJSTRUPDATE ) )
        {
        RISK_ProcessMessage( *ppMessage );

        /* Call the callback for messages ( after changing dist. obj. ) */
        if ( RiskGame.PostViewCallback )
            {
            RiskGame.PostViewCallback( *ppMessage );
            }

        /* Call the replicate message */
        RiskGame.ReplicateCallback( *ppMessage,
                                    MESS_INCOMING,
                                    iSource );

        /* Done with the message */
        NET_DeleteMessage( *ppMessage );

        /* Nobody above here should get wind of this! */
        *ppMessage = NULL;
        }
    else
        {
        /* Call the callback for messages */
        if ( RiskGame.PostViewCallback )
            {
            RiskGame.PostViewCallback( *ppMessage );
            }
        }

    return 1;
    }

/**
 * RISK_SendSyncMessage
 */
Int32 RISK_SendSyncMessage( Int32 iSocket,
                            const Msgs *pMessage,
                            MsgId iReturnMessType,
                            void ( *CBK_MessageReceived )( const Msgs * ) )
    {
    Msgs   *pMess;

    /* Send the message */
    RISK_SendMessage( iSocket, pMessage );

    /* Loop, until we receive the desired message, dispatching
     * all others to the specified callback.
     */

    for ( ; ; )
        {
        /* This will block when there is no input */
        RISK_ReceiveMessage( iSocket, &pMess );

        /* If we received the message we were looking for,
         * then dispatch it and return finally.
         */
        if ( pMess != NULL )
            {
            CBK_MessageReceived( pMess );

            if ( pMess->msgId == iReturnMessType )
                {
                NET_DeleteMessage( pMess );
                return ( 1 );
                }

            NET_DeleteMessage( pMess );
            }
        }
    }


/**
 * RISK_GetNumPlayersOfClient
 * \bug This is a worst case O^2 algorithm.  It could be sped up to
 *      constant time if the number of players was kept track of, but
 *      I doubt this will be much of an overhead.
 */
Int32 RISK_GetNumPlayersOfClient( Int32 iClient )
    {
    Int32   i;
    Int32   iCount;
    Int32   inum;

    assert( iClient >= 0 && iClient < MAX_CLIENTS );

    /* Count up all of the players at this client */
    inum = RISK_GetNumPlayers( );
    for ( i = iCount=0; i < inum; i++ )
        {
        if ( RISK_GetClientOfPlayer( RISK_GetNthPlayer( i )) == iClient )
            {
            iCount++;
            }
        }

    return iCount;
    }


/**
 * RISK_GetNumLivePlayersOfClient
 * \bug This is a worst case O^2 algorithm.  It could be sped up to
 *      constant time if the number of live players was kept track of, but
 *      I doubt this will be much of an overhead.
 */
Int32 RISK_GetNumLivePlayersOfClient( Int32 iClient )
    {
    Int32   i;
    Int32   iCount;

    assert( iClient >= 0 && iClient < MAX_CLIENTS );

    /* Count up all of the players at this client */
    for ( i = iCount=0; i!=RISK_GetNumLivePlayers( ); i++ )
        {
        if ( RISK_GetClientOfPlayer( RISK_GetNthLivePlayer( i )) == iClient )
            {
            iCount++;
            }
        }

    return iCount;
    }


/**
 * RISK_SaveObject
 */
void RISK_SaveObject( CString strFileName )
    {
    UNUSED( strFileName );

    /* I need to write _RISK_[ Read|Write ][ CString|Int32eger ], just like
     * for network.  Use the htonl( ) so that files saved on one
     * endianness can be read on another.  I don't need to save the
     * read-only fields, like the ContinentDatabase stuff.  Since I
     * will save the ID of the field, old game files should work for
     * future version of the game, as long as I don't fiddle with the IDs.
     */
    assert( FALSE );
    }


/**
 * RISK_LoadObject
 */
void RISK_LoadObject( CString strFileName )
    {
    UNUSED( strFileName );

    assert( FALSE );
    }


/**
 * _RISK_GetSpecies
 */
SpeciesObject *_RISK_GetSpecies( Int32 iSpecies )
    {
    SpeciesObject  *ptr = NULL;
    SpeciesObject  *ptrLast;
    Int32           i;

    /* Thread down the linked list as if it were an array,
     * looking for the iSpecies'th object.
     */

    for ( ptrLast = NULL, ptr = &RiskGame.pSpecies, i = 0;
          ptr && i != iSpecies;
          ptr = ptr->next, i++ )
        {
        ptrLast = ptr;
        }

    assert( i == iSpecies || i < iSpecies );

    /* Create a new species object if needed, with
     * blank objects inbetween if needed.
     */

    if ( ptr )
        return ptr;

    /* The current species is NULL, ptrLast points to the last valid one */
    i--;

    for ( ;
          i != iSpecies;
          i++, ptrLast = ptr )
        {
        ptr = ( SpeciesObject * ) MEM_Alloc( sizeof( SpeciesObject ));

        /* Init the object */
        ptr->strName = NULL;
        ptr->strVersion = NULL;
        ptr->strAuthor = NULL;
        ptr->strDescription = NULL;
        ptr->iClient = -1;
        ptr->iAllocationState = ALLOC_NONE;
        ptr->next = NULL;

        /* Append it to the list */
        assert( ptrLast );
        ptrLast->next = ptr;
        }

    assert( ptr );

    return ptr;
    }


/**
 * _RISK_GetIntValue
 */
Int32 _RISK_GetIntValue( Int32 iField, Int32 iIndex1, Int32 iIndex2 )
    {
    SpeciesObject *s;

    switch( iField )
        {
        case PLR_ATTACKMODE:
            return RiskGame.pPlayers[ iIndex1 ].iAttackMode;
            break;

        case PLR_DICEMODE:
            return RiskGame.pPlayers[ iIndex1 ].iDiceMode;
            break;

        case PLR_MSGDSTMODE:
            return RiskGame.pPlayers[ iIndex1 ].iMsgDstMode;
            break;

        case PLR_STATE:
            return RiskGame.pPlayers[ iIndex1 ].iState;
            break;

        case PLR_CLIENT:
            return RiskGame.pPlayers[ iIndex1 ].iClient;
            break;

        case PLR_SPECIES:
            return RiskGame.pPlayers[ iIndex1 ].iSpecies;
            break;

        case PLR_ALLOCATION:
            return RiskGame.pPlayers[ iIndex1 ].iAllocationState;
            break;

        case PLR_NUMCOUNTRIES:
            return RiskGame.pPlayers[ iIndex1 ].iCountriesOwned;
            break;

        case PLR_NUMARMIES:
            return RiskGame.pPlayers[ iIndex1 ].iNumArmies;
            break;

        case PLR_NUMCARDS:
            return RiskGame.pPlayers[ iIndex1 ].iNumCards;
            break;

        case PLR_CARD:
            return RiskGame.pPlayers[ iIndex1 ].piCards[ iIndex2 ];
            break;

        case PLR_MISSION:
            return RiskGame.pPlayers[ iIndex1 ].typOfMission;
            break;

        case PLR_MISSION1:
            return RiskGame.pPlayers[ iIndex1 ].mission1;
            break;

        case PLR_MISSION2:
            return RiskGame.pPlayers[ iIndex1 ].mission2;
            break;

        case CNT_NUMARMIES:
            return RiskGame.pCountries[ iIndex1 ].iNumArmies;
            break;

        case CNT_OWNER:
            return RiskGame.pCountries[ iIndex1 ].piOwner;
            break;

        case SPE_CLIENT:
            s = _RISK_GetSpecies( iIndex1 );
            return s->iClient;
            /* break; */

        case SPE_ALLOCATION:
            s = _RISK_GetSpecies( iIndex1 );
            return s->iAllocationState;
            /* break; */

        default:
            assert( FALSE );
            break;
        }

    return 0;
    }


/**
 * _RISK_GetStrValue
 */
CString _RISK_GetStrValue( Int32 iField, Int32 iIndex1, Int32 iIndex2 )
    {
    SpeciesObject *s;

    UNUSED( iIndex2 );

    switch( iField )
        {
        case PLR_NAME:
            return RiskGame.pPlayers[ iIndex1 ].strName;
            /* break; */

        case PLR_COLORSTRING:
            return RiskGame.pPlayers[ iIndex1 ].strColor;
            /* break; */

        case SPE_NAME:
            s = _RISK_GetSpecies( iIndex1 );
            return s->strName;
            /* break; */
        case SPE_VERSION:
            s = _RISK_GetSpecies( iIndex1 );
            return s->strVersion;
            /* break; */

        case SPE_DESCRIPTION:
            s = _RISK_GetSpecies( iIndex1 );
            return s->strDescription;
            /* break; */

        case SPE_AUTHOR:
            s = _RISK_GetSpecies( iIndex1 );
            return s->strAuthor;
            /* break; */

        default:
            assert( FALSE );
            break;
        }

    return NULL;
    }


/**
 * Replicate species and player registration data to new client.
 */
void RISK_ReplicateRegistrationData( Int32 iCommLinkDest )
    {
    Int32 i;
    Int32 iPlayer;
    Int32 iSpecies;

    /* Ignore the "human" entry... */
    for ( i = iSpecies = 1; i < RISK_GetNumSpecies( ); i++, iSpecies++ )
        {
        while ( RISK_GetAllocationStateOfSpecies( iSpecies ) == ALLOC_NONE )
            {
            iSpecies++;
            }

        RISK_SelectiveReplicate( iCommLinkDest, SPE_NAME, iSpecies, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, SPE_VERSION, iSpecies, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, SPE_DESCRIPTION, iSpecies, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, SPE_AUTHOR, iSpecies, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, SPE_CLIENT, iSpecies, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, SPE_ALLOCATION, iSpecies, 0 );
        }

    for ( i = 0; i < RISK_GetNumPlayers( ); i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );

        RISK_SelectiveReplicate( iCommLinkDest, PLR_STATE, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_COLORSTRING, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_NAME, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_SPECIES, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_CLIENT, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_ALLOCATION, iPlayer, 0 );
        }
    }

/**
 * Replicates all game state data to new client.
 */
void RISK_ReplicateAllData( Int32 iCommLinkDest )
    {
    Int32 i;
    Int32 j;
    Int32 iPlayer;

    /* First replicate the registration data */
    RISK_ReplicateRegistrationData( iCommLinkDest );

    /* Now everything else */
    for ( i = 0; i < RISK_GetNumPlayers( ); i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );

        RISK_SelectiveReplicate( iCommLinkDest, PLR_NUMCOUNTRIES, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_NUMARMIES, iPlayer, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, PLR_NUMCARDS, iPlayer, 0 );

        for ( j = 0; j != RISK_GetNumCardsOfPlayer( iPlayer ); j++ )
            {
            RISK_SelectiveReplicate( iCommLinkDest, PLR_CARD, iPlayer, j );
            }
        }

    for ( i = 0; i != NUM_COUNTRIES; i++ )
        {
        RISK_SelectiveReplicate( iCommLinkDest, CNT_OWNER, i, 0 );
        RISK_SelectiveReplicate( iCommLinkDest, CNT_NUMARMIES, i, 0 );
        }
    }

/**
 * Handles allocation of new player.
 */
void RISK_HandleALLOCPLAYER( Int32 iClient, Int32 iPlayer )
    {
    MsgReplyPacket  mess;

    /* Get a player and return it, or -1 if none available */
    mess.iReply = iPlayer;

    /* If there was a valid player, make a note */
    if ( iPlayer != -1 )
        {
        /* Reset all of the player fields */
        RISK_SetAttackModeOfPlayer( iPlayer, 1 );
        RISK_SetStateOfPlayer( iPlayer, PLAYER_ALIVE );
        RISK_SetClientOfPlayer( iPlayer, -1 );
        RISK_SetNumCountriesOfPlayer( iPlayer, 0 );
        RISK_SetNumArmiesOfPlayer( iPlayer, 0 );
        RISK_SetNumCardsOfPlayer( iPlayer, 0 );
        RISK_SetMissionTypeOfPlayer( iPlayer, NO_MISSION );

        /* Note that the player is in the process of being allocated.
         * The client will complete the procedure.  The client that
         * sets the allocation state to ALLOC_COMPLETE is the one
         * responsible for increasing the number of players.
         */

        RISK_SetAllocationStateOfPlayer( iPlayer, ALLOC_INPROGRESS );
        }

    mess.msgId = MSG_REPLYPACKET;
    RISK_SendMessage( CLIENTS_GetCommLinkOfClient( iClient ),
                      ( const Msgs * ) &mess );
    }

/* End of file */
