/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: server.c,v 1.54 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * Server main loop and associated functionality.
 */

static const char
    rcsid[] = "@(#)$Id: server.c,v 1.54 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include "language.h"
#include "types.h"
#include "game.h"
#include "riskgame.h"
#include "network.h"
#include "server.h"
#include "deck.h"
#include "debug.h"
#include "clients.h"

/*
 * Server data
 */
static Int32      iServerCommLink;                  /**< Server socket */
static fd_set     fdSet;                            /**< Client socket states */
static fd_set     fdBackup;                         /**< Client socket states backup */
static Int32      iMaxFileDescUsed = -1;            /**< ??? */

/*
 * Game data
 */
static Int32      iState;                           /**< */
static Deck      *pPlayerDeck;                      /**< Deck of free player ids */
static Deck      *pCardDeck;                        /**< Card deck */
static Int32      iReply;                           /**< ??? */
static Flag       fGameReset = TRUE;                /**< ??? */
static Flag       fRememberKilled = FALSE;          /**< ??? */
static Int32      iTurn;                            /**< ??? */
static Int32      iFirstPlayer;                     /**< ??? */



/* Private functions */
void        SRV_ResetGame( void );
void        SRV_AttemptNewGame( void );
void        SRV_DistributeCountries( void );
Flag        SRV_DistributeMissions( void );
void        SRV_SetInitialArmiesOfPlayers( void );
void        SRV_SetInitialMissionOfPlayers( void );
void        SRV_NotifyClientsOfTurn( Int32 iTurn );
void        SRV_HandleRegistration( Int32 iCommLink );
int         SRV_HandleMessage( Int32 iClient, Msgs *pMessage );
void        UTIL_ExitProgram( Int32 iExitValue );
Int32       SRV_IterateTurn( void );
RETSIGTYPE  SRV_HandleSignals( Int32 iParam );

/* Server message handlers */
void        SRV_HandleEXIT( Int32 iExitValue );
void        SRV_HandleFREEPLAYER( const Msgs *pMessage );
void        SRV_HandleREPLYPACKET( const Msgs *pMessage );
void        SRV_HandleMESSAGEPACKET( Int32 iClient, const Msgs *pMessage );
void        SRV_HandleENTERSTATE( const Msgs *pMessage );
void        SRV_HandleDEREGISTERCLIENT( Int32 iClient );


/**
 * Server initialization ( signal handling, creating server socket, ... )
 */
void SRV_Init( void )
    {
    struct sockaddr_in   server;
    struct timeval tv;

    /* Print an informative message */
    printf( "%s: %s %s\n", SERVERNAME, SERVER_STARTING , VERSION );

    /* Catch signals so that we can clean up upon termination */
    signal( SIGHUP,  SRV_HandleSignals );
    signal( SIGINT,  SRV_HandleSignals );
    signal( SIGQUIT, SRV_HandleSignals );
    signal( SIGILL,  SRV_HandleSignals );
    signal( SIGTRAP, SRV_HandleSignals );
    signal( SIGFPE,  SRV_HandleSignals );
    signal( SIGKILL, SRV_HandleSignals );
    signal( SIGBUS,  SRV_HandleSignals );
    signal( SIGTERM, SRV_HandleSignals );

    /* We want to ignore this signal, because we don't want a I/O
     * failure to terminate the program.
     */

    signal( SIGPIPE, SIG_IGN );

    /* Init. clients and players */
    CLIENTS_Init( );

    /* Initialize the random generator */
    gettimeofday( &tv, NULL );
    srand( tv.tv_sec *
           tv.tv_usec +
           ( getppid( ) *
             getpid( ) *
             getsid( getpid( ) ) ) );

    /* Initialize the pool of free players */
    pPlayerDeck = DECK_Create( MAX_PLAYERS );

    /* Create server socket */
    if ( ( iServerCommLink = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
        {
        perror( "Creating CommLink" );
        UTIL_ExitProgram( 1 );
        }

    NET_SetCommLinkOptions ( iServerCommLink );

    /* Update the max */
    iMaxFileDescUsed = MAX( iMaxFileDescUsed, iServerCommLink );

    /* Name sockets using wildcards */
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons( RISK_PORT );

    /* Apply socket options to the socket before binding */
    NET_SetCommLinkOptions( iServerCommLink );

    /* Bind the socket to the port */
    if ( bind( iServerCommLink, ( struct sockaddr * ) &server, sizeof( server ) ) )
        {
        printf( "%s: Xfrisk port %d %s\n", SERVERNAME, RISK_PORT, ERR_PORT_IN_USE );
        UTIL_ExitProgram( 1 );
        }
    }


/**
 * Broadcasts the given textual message to all clients
 */
void SRV_BroadcastTextMessage( CString strMessage )
    {
    MsgMessagePacket   msgMess;

    msgMess.strMessage  = strMessage;
    msgMess.iFrom       = FROM_SERVER;
    msgMess.iTo         = DST_ALLPLAYERS;
    msgMess.msgId       = MSG_MESSAGEPACKET;

    /* Send the message out to all clients */
    SRV_BroadcastMessage( ( const Msgs * ) &msgMess );
    }


/**
 * Main server message handler. Branches to message specific handlers.
 */
int SRV_HandleMessage( Int32 iClient, Msgs *pMessage )
    {
    char    buf[ 256 ];
    int     handled = FALSE;
    Int32   iPlayer;
    Msgs    msg;
    Int32   i;
    Flag    fFoundPlayer;

    /* Depending on the message type, dispatch it to one
     * of the handlers.  They usually take just a message,
     * but depending on what they do they could take the
     * and/or the client index.
     */
    assert( iClient <= MAX_CLIENTS );

    switch ( iGameState )
        {
        case GAME_REGISTERING:
            switch ( pMessage->msgId )
                {
                case MSG_STARTGAME:
                    CLIENTS_SetStartState( iClient, TRUE );
                    /* Informative message */
                    /* should tell which player, not easy*/
                    printf( "%s: %s %s\n",
                            SERVERNAME,
                            CLIENTS_GetAddress( iClient ),
                            FINISHED_REGISTRY );
                    snprintf( buf,
                              sizeof( buf ),
                              "%s: %s %s",
                              SERVERNAME,
                              NEW_CLIENT,
                              FINISHED_REGISTRY );
                    SRV_BroadcastTextMessage( buf );

                    /* Reset the game, and possible start it */
                    /* Will only reset if not already done... */
                    SRV_ResetGame( );
                    /* The first AI-Client can't start the game */
                    if ( ( CLIENTS_GetType( iClient ) != CLIENT_AI )
                         || ( CLIENTS_GetNumClientsStarted( ) > 1 ) )
                        {
                        SRV_AttemptNewGame( );
                        }
                    handled = TRUE;
                    break;

                default:
                    break;
                }
            break;

        case GAME_FORTIFYING:
            switch ( pMessage->msgId )
                {
                case MSG_ENDTURN:
                    /* Sanity check */
                    assert( RISK_GetNumLivePlayers( ) );

                    /* If noone has any more armies, then move to
                     * GAME_PLAYING mode.  Otherwise, go along
                     * the list of players, looking for a player who
                     * still has armies to place.
                     */

                    for ( i = 0, fFoundPlayer = FALSE;
                          ( i != RISK_GetNumLivePlayers( ) ) && !fFoundPlayer;
                          i++ )
                        {
                        /* The next player who's turn it is */
                        iPlayer = SRV_IterateTurn( );

                        /* Is there a player with armies left? */
                        if ( RISK_GetNumArmiesOfPlayer( iPlayer ) != 0 )
                            {
                            SRV_NotifyClientsOfTurn( iPlayer );
                            fFoundPlayer = TRUE;
                            }
                        }

                    /* If we have not found a player with armies,
                     * then there are no longer any players with any
                     * armies to fortify with.  Let the game begin!!
                     * Theorem ( proof left to reader ):  When some
                     * players have more armies to to fortify with
                     * than others ( because they got less countries ),
                     * then the last player to fortify here will be
                     * the last player in relation to the first player
                     * who fortified.  So if we move to the next
                     * player, it will be the player who fortified
                     * first, which is what we want.
                     *
                     * This theorem is false when you use bpk's
                     * multiple-fortify hack. See my comment in
                     * SRV_AttemptNewGame. --Pac.
                     */

                    if ( !fFoundPlayer )
                        {
                        iGameState = GAME_PLAYING;
                        iTurn = iFirstPlayer;
                        iPlayer = iFirstPlayer;
                        SRV_NotifyClientsOfTurn( iPlayer );
                        }
                    handled = TRUE;
                    break;

                default:
                    break;
                }
            break;

        default:
            switch ( pMessage->msgId )
                {
                case MSG_STARTGAME:
                    CLIENTS_SetStartState( iClient, TRUE );
                    /* Some random client tried to start in the middle
                     * of a game, probably an overeager AI client.
                     */
                    SRV_BroadcastTextMessage( "somebody tried to join after game was started" );
                    printf( "client %s tried to join after game started\n",
                            CLIENTS_GetAddress( iClient ) );
                    handled = TRUE;
                    break;

                case MSG_ENDTURN:
                    /* Get the next player */
                    iPlayer = SRV_IterateTurn( );
                    SRV_NotifyClientsOfTurn( iPlayer );
                    handled = TRUE;
                    break;

                default:
                    break;
                }
            break;
        }

    /*
     * Messages with identical handling in all states
     */
    if ( !handled )
        {
        switch ( pMessage->msgId )
            {
            case MSG_NETMESSAGE:
                /* fall-through */
            case MSG_DICEROLL:
                /* fall-through */
            case MSG_PLACENOTIFY:
                /* fall-through */
            case MSG_ATTACKNOTIFY:
                /* fall-through */
            case MSG_MOVENOTIFY:
                SRV_CompleteBroadcast( iClient, pMessage );
                break;
            case MSG_ENDOFMISSION:
                /* fall-through */
            case MSG_VICTORY:
                SRV_BroadcastMessage( pMessage );
                break;
            case MSG_MISSION:
                if ( SRV_DistributeMissions( ) )
                    {
                    SRV_BroadcastMessage( pMessage );
                    }
                break;

            case MSG_ENDOFGAME:
                iGameState = GAME_REGISTERING;
                /* Let everyone else know */
                msg.msgId = MSG_ENDOFGAME;
                SRV_BroadcastMessage( &msg );
                SRV_ResetGame( );
                break;

            case MSG_ALLOCPLAYER:
                RISK_HandleALLOCPLAYER( iClient, DECK_GetCard( pPlayerDeck ) );
                break;

            case MSG_FREEPLAYER:
                SRV_HandleFREEPLAYER( pMessage );
                break;

            case MSG_REPLYPACKET:
                SRV_HandleREPLYPACKET( pMessage );
                break;

            case MSG_MESSAGEPACKET:
                SRV_HandleMESSAGEPACKET( iClient, pMessage );
                break;

            case MSG_DEREGISTERCLIENT:
                /* Do the actual deregistration */
                SRV_HandleDEREGISTERCLIENT( iClient );

                /* Informative message */
                sprintf( buf,
                         "%s: %s %s\n",
                         SERVERNAME,
                         CLIENTS_GetAddress( iClient ),
                         HAS_DEREGISTERED );
                printf( "%s", buf );
                SRV_BroadcastTextMessage( "A client has deregistered" );
                break;

            case MSG_ENTERSTATE:
                SRV_HandleENTERSTATE( pMessage );
                break;

            case MSG_EXCHANGECARDS:
            {
            MsgExchangeCards *pMess = &( pMessage->msgExchangeCards );
            MsgReplyPacket    msgMess;

            msgMess.iReply = DECK_ExchangeCards( pCardDeck, pMess->piCards );
            msgMess.msgId = MSG_REPLYPACKET;
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( iClient ),
                              ( const Msgs * ) &msgMess );
            }
            break;

            case MSG_REQUESTCARD:
            {
            MsgRequestCard *pMess = &( pMessage->msgRequestCard );

            RISK_SetCardOfPlayer( pMess->iPlayer,
                                  RISK_GetNumCardsOfPlayer( pMess->iPlayer ),
                                  DECK_GetCard( pCardDeck ) );
            RISK_SetNumCardsOfPlayer( pMess->iPlayer,
                                      ( RISK_GetNumCardsOfPlayer( pMess->iPlayer ) +
                                        1 ) );
            }
            break;

            case MSG_FORCEEXCHANGECARDS:
            {
            Int32 iPlayer;

            iPlayer = pMessage ->msgForceExchangeCards.iPlayer;
            if ( RISK_GetNumCardsOfPlayer( iPlayer ) < 5 )
                {
                pMessage->msgForceExchangeCards.iPlayer = -1;
                }
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( iClient ),
                              pMessage );
            }
            break;

            case MSG_EXIT:
                SRV_HandleEXIT( 0 );
                break;

            case MSG_NOMESSAGE:
                break;

            default:
            {
            MsgNetPopup popupmsg;

            /* Assume that client is messed up.  Consider it
             * a failure and kill the client.
             */
            printf( "%s: %s %s: %s",
                    SERVERNAME,
                    CLIENT,
                    CLIENT_DEAD,
                    INVALID_MESSAGE );
            popupmsg.strMessage = buf;
            popupmsg.msgId = MSG_NETPOPUP;
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( iClient ),
                              ( const Msgs * ) &popupmsg );

            /* Log the failure */
            SRV_LogFailure( "Sent bogus message",
                            CLIENTS_GetCommLinkOfClient( iClient ) );
            }

            /* Free up the memory the message was taking */
            NET_DeleteMessage( pMessage );
            }
        }

    return( 0 );
    }


/**
 * Server main loop - 'Play the game'
 */
void SRV_PlayGame( void )
    {
    Int32   iClient;
    Int32   iError;
    Int32   iRetVal;
    Msgs    *pMessage;
    Msgs    msg;

    /* Create the card deck */
    pCardDeck = DECK_Create( NUM_COUNTRIES + 2 );

    /* Add the initial fd to keep an eye on -- the connect socket */
    FD_ZERO( &fdBackup );
    FD_SET( iServerCommLink, &fdBackup );

    /* Start accepting connections */
    listen( iServerCommLink, 5 );

    /* Loop for the entirety of the game */
    for ( ; ; )
        {
        fdSet = fdBackup;

        /* If there have been any failures, then deal with them */
        SRV_RecoverFailures( );

        /* Wait for a message to come in */
        if ( select( iMaxFileDescUsed + 1,
                     ( fd_set * ) &fdSet,
                     ( fd_set * ) 0,
                     ( fd_set * ) 0,
                     NULL ) < 0 )
            {
            /* errno = EINTR is caused by SRV_HandleSignals */
            if ( errno != EINTR )
                {
                perror( "Select" );
                }
            }

        /* Two things might have happened here.  Either an existing client
         * sent a message to the server, or a new client sent a message to
         * the connect port, trying to join the game.  If the former occurred
         * process it normally.  If the latter occurred, let the client
         * connect, and add its fd to the fd map.  If we are in the middle
         * of a game, send it a message saying this, and then perform a
         * RISK_SelectiveReplicate( ) so as to get the new client in the
         * same state as the others.  Also send a message to the other
         * clients telling them what is happening.
         */

        if ( FD_ISSET( iServerCommLink, &fdSet ) )
            { /* new client trying to connect */
            Int32 iNewCommLink;

            /* Try to accept the new connection */
            iNewCommLink = accept( iServerCommLink, 0, 0 );
            if ( iNewCommLink < 0 )
                {
                /* Couldn't do it, go back to top */
                printf( "%s %s\n", SERVERNAME, SERVER_CONNECT_FAILED );
                continue;
                }
            else
                {
                /* connection went fine */
                /* Does Xfrisk have enough resources to hold this client? */
                if ( ( CLIENTS_GetNumClients( ) >= MAX_CLIENTS )
                     || ( iNewCommLink >= MAX_DESCRIPTORS ) )
                    {
                    msg.msgId = MSG_EXIT;
                    RISK_SendMessage( iNewCommLink, &msg );
                    close( iNewCommLink );
                    continue;
                    }
                }

            /* Assert: At this point, connection is complete and we have
             * enough resources to keep it around.  Begin connection protocol.
             */

            /* Set options on new CommLink */
            NET_SetCommLinkOptions( iNewCommLink );
            SRV_HandleRegistration( iNewCommLink );
            }
        else
            {
            /* no FD_ISSET,   there is a client trying to send a message */
            for ( iClient = 0; iClient < MAX_CLIENTS ; iClient++ )
                {
                if ( ( CLIENTS_GetAllocationState( iClient ) == ALLOC_COMPLETE )
                     && ( FD_ISSET( CLIENTS_GetCommLinkOfClient( iClient ), &fdSet ) ) )
                    {
                    iRetVal = RISK_ReceiveMessage( CLIENTS_GetCommLinkOfClient( iClient ),
                                                   &pMessage );
                    if ( !iRetVal && pMessage == NULL )
                        {
                        SRV_LogFailure( "Socket read",
                                        CLIENTS_GetCommLinkOfClient( iClient ) );
                        continue;
                        }
                    else if ( !iRetVal || pMessage == NULL )
                        {
                        continue;
                        }
                    iError = SRV_HandleMessage( iClient, pMessage );
                    if ( iError != 0 )
                        {
                        printf( "error %d when handling messagetype %d from %s\n",
                                iError,
                                pMessage->msgId,
                                CLIENTS_GetAddress( iClient ) );
                        }
                    break;
                    }
                }
            }
        }
    }


/**
 * Notify all clients of turn change
 */
void SRV_NotifyClientsOfTurn( Int32 iPlayer )
    {
    MsgTurnNotify msgTurnNotify;

    /* Let all clients know whos turn it is */
    msgTurnNotify.iPlayer = iPlayer;
    msgTurnNotify.iClient = RISK_GetClientOfPlayer( iPlayer );
    msgTurnNotify.msgId = MSG_TURNNOTIFY;
    SRV_BroadcastMessage( ( const Msgs * ) &msgTurnNotify );
    }


/**
 * Distribute countries among live players.
 */
void SRV_DistributeCountries( void )
    {
    Deck   *pCountryDeck;
    Int32   iCountry;
    Int32   iPlayer;
    Int32   i;

    pCountryDeck = DECK_Create( NUM_COUNTRIES );

    /* Dole out the countries */
    for ( i = 0; i != NUM_COUNTRIES; i++ )
        {
        /* Pick a country, any country */
        iCountry = DECK_GetCard( pCountryDeck );

#ifdef TEST_GAME
        /* Give countries to the first player, leave one for the rest */
        if ( i <= ( NUM_COUNTRIES - RISK_GetNumLivePlayers( ) ) )
            {
            iPlayer = RISK_GetNthLivePlayer( 0 );
            }
        else
            {
            iPlayer =
                RISK_GetNthLivePlayer( i - ( NUM_COUNTRIES
                                             - RISK_GetNumLivePlayers( ) ) );
            }
#else
        iPlayer = iTurn;
#endif

        /* Update the game object */
        RISK_SetOwnerOfCountry( iCountry, iPlayer );
        RISK_SetNumCountriesOfPlayer( iPlayer,
                                      RISK_GetNumCountriesOfPlayer(
                                          iPlayer ) + 1 );
        RISK_SetNumArmiesOfCountry( iCountry, 1 );
        RISK_SetNumArmiesOfPlayer( iPlayer,
                                   RISK_GetNumArmiesOfPlayer( iPlayer ) - 1 );

        /* Iterate to next player */
        SRV_IterateTurn( );
        }

#ifdef TEST_GAME
    /* Set the number of armies to be small, to let the game start soon. */
    for ( i = 0; i != RISK_GetNumLivePlayers( ); i++ )
        {
        RISK_SetNumArmiesOfPlayer( RISK_GetNthLivePlayer( i ), 2 );
        }
#endif

    DECK_Destroy( pCountryDeck );
    }


/**
 * Calculate number of armies given to each player at game start
 * \bug Part of rules is hardcoded here
 */
void SRV_SetInitialArmiesOfPlayers( void )
    {
    Int32       i;
    Int32       iPlayer;
    Int32       iNumArmies;
    const Int32 iNumPlayers = RISK_GetNumPlayers( );

    assert( iNumPlayers >= 2 );

    /* Calculate the number of armies. */
    iNumArmies = MAX( 50 - iNumPlayers * 5, /* According to the rules */
                      NUM_COUNTRIES / iNumPlayers + 1 ); /* At least one per country */

    /* Make sure it's enough */
    assert( ( iNumPlayers * iNumArmies ) >= NUM_COUNTRIES );

    /* Set the initial number of armies for all the players */
    for ( i = 0; i < RISK_GetNumPlayers( ); i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );
        RISK_SetNumArmiesOfPlayer( iPlayer, iNumArmies );

        /* Sanity check */
        assert( RISK_GetNumArmiesOfPlayer( iPlayer ) > 0 );
        }
    }


/**
 * Set initial mission of all players to "no mission"
 */
void SRV_SetInitialMissionOfPlayers( void )
    {
    Int32 i;
    Int32 iPlayer;
    Int32 nb;

    nb = RISK_GetNumPlayers( );
    /* Set the initial mission's type for all the players */
    for ( i = 0; i < nb; i++ )
        {
        iPlayer = RISK_GetNthPlayer( i );
        RISK_SetMissionTypeOfPlayer( iPlayer, NO_MISSION );
        }
    }


/**
 * ????????
 * \bug ( MSH 25.12.99 ) Obscure code with no comments!
 */
Flag SRV_DistributeMissions( void )
    {
    Int32 iMission[ MAX_PLAYERS ];
    Int32 iPlayer;
    Int32 i;
    Int32 j;
    Int32 cont1;
    Int32 cont2;
    Int32 n;
    Int32 nb;
    Int32 m;

    nb = RISK_GetNumLivePlayers( );

    for ( n = 0; n < nb; n++ )
        {
        iPlayer = RISK_GetNthPlayer( n );
        if ( RISK_GetMissionTypeOfPlayer( iPlayer ) != NO_MISSION )
            {
            return FALSE;
            }
        }

    for ( n = 0; n < MAX_PLAYERS; n++ )
        {
        iMission[ n ] = -1;
        }

    for ( n = 0; n < nb; n++ )
        {
        iPlayer = RISK_GetNthLivePlayer( n );
        i = rand( ) % ( 2 + ( NUM_CONTINENTS * ( NUM_CONTINENTS - 1 ) ) / 2 + nb - 1 );
        j = 0;
        while ( j < MAX_PLAYERS )
            {
            if ( i == iMission[j] )
                {
                j = 0;
                i = ( i + 1 ) %( 2 + NUM_CONTINENTS * ( NUM_CONTINENTS - 1 )
                                 + nb - 1 );
                }
            else
                j++;
            }
        iMission[ iPlayer ] = i;
        if ( i == 0 )
            {
            RISK_SetMissionTypeOfPlayer( iPlayer, CONQUIER_WORLD );
            }
        else if ( i == 1 )
            {
            RISK_SetMissionTypeOfPlayer( iPlayer, CONQUIER_Nb_COUNTRY );
            RISK_SetMissionNumberOfPlayer( iPlayer, NUM_COUNTRIES / 2 );
            }
        else if ( i <= ( ( NUM_CONTINENTS * ( NUM_CONTINENTS - 1 ) ) / 2 + 1 ) )
            {
            i = i - 2;
            cont1 = -1;
            m = 0;
            while ( cont1 == -1 )
                {
                cont2 = m + i + 1;
                if ( cont2 < NUM_CONTINENTS )
                    {
                    cont1 = m;
                    }
                i = i - ( NUM_CONTINENTS - m - 1 );
                m++;
                }
            RISK_SetMissionTypeOfPlayer( iPlayer, CONQUIER_TWO_CONTINENTS );
            RISK_SetMissionContinent1OfPlayer( iPlayer, cont1 );
            RISK_SetMissionContinent2OfPlayer( iPlayer, cont2 );
            }
        else
            {
            i = i - ( NUM_CONTINENTS * ( NUM_CONTINENTS - 1 ) ) / 2 - 2;
            j = RISK_GetNthLivePlayer( i );
            if ( j == iPlayer )
                {
                i++;
                i = i % nb;
                j = RISK_GetNthLivePlayer( i );
                }
            RISK_SetMissionTypeOfPlayer( iPlayer, KILL_A_PLAYER );
            RISK_SetMissionMissionPlayerToKillOfPlayer( iPlayer, j );
            RISK_SetMissionPlayerIsKilledOfPlayer( iPlayer, FALSE );
            }
        }
    return TRUE;
    }


/**
 * Broadcast given message to all active clients.
 */
void SRV_BroadcastMessage( const Msgs *pMessage )
    {
    Int32 i;

    /* If there are no clients, then do nothing */
    if ( CLIENTS_GetNumClients( ) == 0 )
        {
        return;
        }

    /* Loop through and send the message to each active client */
    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        if ( CLIENTS_GetAllocationState( i ) == ALLOC_COMPLETE )
            {
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( i ),
                              pMessage );
            }
        }
    }


/**
 * Broadcast given message to all active clients except originator.
 */
void SRV_CompleteBroadcast( Int32 iClientExclude,
                            const Msgs *pMess )
    {
    Int32 i;

    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        if ( ( i != iClientExclude )
             && ( CLIENTS_GetAllocationState( i ) == ALLOC_COMPLETE ) )
            {
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( i ),
                              pMess );
            }
        }
    }


/**
 * Send message to all clients. Handles server-originated messages and
 * relaying of client-originated messages.
 *
 * \b History:
 * \arg 08.18.94  ESF  Created.
 */
void SRV_Replicate( const Msgs *pMess, Int32 iDirection, Int32 iSrc )
    {
    if ( iDirection == MESS_OUTGOING )
        {
        SRV_BroadcastMessage( pMess );
        }
    else /* ( iType == MESS_INCOMING ) */
        {
        /* Convert the CommLink to a client */
        SRV_CompleteBroadcast( CLIENTS_GetCommLinkToClient( iSrc ),
                               pMess );
        }
    }


/**
 * Server exit cleanup.
 */
void SRV_HandleEXIT( Int32 iExitValue )
    {
    Int32   n;
    Msgs    msg;

    /* close the listening socket */
    close( iServerCommLink );

    /* close all of the player sockets */
    for ( n = 0; n != MAX_CLIENTS; n++ )
        {
        if ( CLIENTS_GetAllocationState( n ) == ALLOC_COMPLETE )
            {
            msg.msgId = MSG_EXIT;
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( n ),
                              &msg );
            close( CLIENTS_GetCommLinkOfClient( n ) );
            }
        }

    UTIL_ExitProgram( iExitValue );
    }


/**
 * \b History:
 * \arg 06.10.94  ESF  Created.
 */
void SRV_HandleREPLYPACKET( const Msgs *pMessage )
    {
    iReply = pMessage->msgReplyPacket.iReply;
    }


/**
 * Handles player resignation during game.
 * \bug Terminates the game - recovery procedures needed!
 */
void SRV_HandleFREEPLAYER( const Msgs *pMessage )
    {
    Int32   iPlayer;
    Int32   i;
    char    buf[ 256 ];
    Msgs    msg;

    /* Put the player ID back onto pool of free players */
    iPlayer = pMessage->msgFreePlayer.iPlayer;
    DECK_PutCard( pPlayerDeck, iPlayer );

    /* If the player has cards at this point, take them away and put
     * them back onto the card deck.
     */

    for ( i = 0; i != RISK_GetNumCardsOfPlayer( iPlayer ); i++ )
        {
        DECK_PutCard( pCardDeck, RISK_GetCardOfPlayer( iPlayer, i ) );
        }

    /* Details, details... */
    RISK_SetAllocationStateOfPlayer( iPlayer, ALLOC_NONE );

    /* If there were no live players at the client then we can let it go
     * without ending the game.  Otherwise, end the game if we're in the
     * PLAY or FORTIFICATION stages of the game.  If the number of live
     * players went down to 0, then we must go back to registering mode.
     * Kind of moot, but safe...
     */

    /* Was the player alive? */
    if ( ( RISK_GetStateOfPlayer( iPlayer ) == PLAYER_ALIVE )
         && ( ( iGameState == GAME_PLAYING )
              || ( iGameState == GAME_FORTIFYING ) ) )
        {
        /* Informative message */
        snprintf( buf, sizeof( buf ), GAME_OVER );
        msg.msgNetPopup.strMessage = buf;
        msg.msgNetPopup.msgId = MSG_NETPOPUP;
        SRV_BroadcastMessage( &msg );

        /* There was no winner, reset the game */
        msg.msgId = MSG_ENDOFGAME;
        SRV_BroadcastMessage( &msg );

        /* It was the end of the game, so reset things */
        SRV_ResetGame( );
        }
    else if ( RISK_GetNumLivePlayers( ) == 0 )
        {
        SRV_ResetGame( );
        }
    }


/**
 * Sends message from client to all clients, all but originator or
 * specified recipient.
 */
void SRV_HandleMESSAGEPACKET( Int32 iClient, const Msgs *pMessage )
    {
    if ( pMessage->msgMessagePacket.iTo == DST_ALLPLAYERS )
        {
        SRV_BroadcastMessage( pMessage );
        }
    else if ( pMessage->msgMessagePacket.iTo == DST_ALLBUTME )
        {
        SRV_CompleteBroadcast( iClient, pMessage );
        }
    else
        {
        RISK_SendMessage( CLIENTS_GetCommLinkOfClient(
                              RISK_GetClientOfPlayer( pMessage->msgMessagePacket.iTo ) ),
                          pMessage );
        }
    }


/**
 * Saves distributed server object state received from client.
 * \bug ( MSH 25.12.99 ) Will figure out a better way to do this.
 */
void SRV_HandleENTERSTATE( const Msgs *pMessage )
    {
    iState = pMessage->msgEnterState.iState;
    }


/**
 * Helper macro for cleaning up aborted registration process.
 */
#define _SRV_CleanupRegistration( )                 \
    {                                               \
    /* NET_DeleteMessage( iMessType, pvMess ); */   \
    FD_CLR( iCommLink, &fdBackup );                 \
    close( iCommLink );                             \
    }


/**
 * Handles registration of a new client.
 * \b Note: contains code for connection re-estabilishment ( ??? )
 */
void SRV_HandleRegistration( Int32 iCommLink )
    {
    MsgClientIdent      msgClientIdent;
    MsgVersion          msgVersion;
    MsgMessagePacket    msgMess;
    Msgs                msg;
    Msgs               *pMess = NULL;
    Msgs               *pDeleteMe = NULL;
    Int32               iNewClient;
    Int32               iClientType = CLIENT_NORMAL;
    Int32               iClientSpecies = SPECIES_HUMAN;
    Flag                fOldClient = FALSE;
    CString             strClientAddress = NULL;
    char                buf[ 256 ];

    /* The Protocol:
     *
     *                      Receive MSG_HELLO           --> OK!
     *                              Other               --> Protocol bogosity.
     *  Send MSG_VERSION.
     *                      Receive MSG_REGISTERCLIENT  --> OK!
     *                              MSG_VERSION         --> Version mismatch.
     *                              Other               --> Protocol bogosity.
     *  Send MSG_CLIENTIDENT if server is not full.
     *  Send MSG_CLIENTEXIT  if server is full.
     */

    /* It should be a MSG_HELLO! */
    if ( !RISK_ReceiveMessage( iCommLink, &pMess )
         || ( ( pMess->msgId != MSG_HELLO )
              && ( pMess->msgId != MSG_OLDREGISTERCLIENT ) ) )
        {
        printf( "%s: %s %d %s\n",
                SERVERNAME,
                ERROR_PROTOCOL,
                pMess ? pMess->msgId : 0,
                IGNORE_CLIENT );
        _SRV_CleanupRegistration( );
        return;
        }

    /* See if it's an old client */
    fOldClient = ( pMess->msgId == MSG_OLDREGISTERCLIENT );

    /* If it is an old client, then get its name */
    if ( fOldClient )
        {
        strClientAddress = pMess->msgOldRegisterClient.strClientAddress;

        /* Mark this message for future deletion */
        pDeleteMe = pMess;
        }
    else
        {
        NET_DeleteMessage( pMess );
        }

    /* If we're talking to a new client, send the version information */
    if ( !fOldClient )
        {
        /* Send the version */
        msgVersion.strVersion = VERSION;
        msgVersion.msgId = MSG_VERSION;
        RISK_SendMessage( iCommLink, ( const Msgs * ) &msgVersion );

        /* It should be a MSG_REGISTERCLIENT or a MSG_VERSION! */
        if ( !RISK_ReceiveMessage( iCommLink, &pMess )
             || ( pMess->msgId != MSG_REGISTERCLIENT ) )
            {
            if ( pMess->msgId == MSG_VERSION )
                {
                printf( "%s: %s (%s)\n",
                        SERVERNAME,
                        CLIENT_MISMATCH,
                        pMess->msgVersion.strVersion );
                }
            else
                {
                printf( "%s: %s (%d) %s\n",
                        SERVERNAME,
                        ERROR_PROTOCOL,
                        pMess->msgId,
                        IGNORE_CLIENT );
                }
            _SRV_CleanupRegistration( );
            return;
            }

        /* Now we know the name ( and type ) of the client */
        strClientAddress = pMess->msgRegisterClient.strClientAddress;
        iClientType      = pMess->msgRegisterClient.iClientType;
        }

    /* If the client is an "AI" client, then send it the
     * dynamic ID of its species, which we allocate on-the-fly.
     */

    if ( iClientType == CLIENT_AI )
        {
        msg.msgSpeciesIdent.iSpeciesID = SRV_AllocSpecies( );
        iClientSpecies = msg.msgSpeciesIdent.iSpeciesID;
        msg.msgSpeciesIdent.msgId = MSG_SPECIESIDENT;
        RISK_SendMessage( iCommLink, &msg );
        }

    /* Get the handle to a new client */
    iNewClient = CLIENTS_Alloc( );

    /* Did the allocation fail? */
    if ( iNewClient == -1 )
        {
        msg.msgId = MSG_EXIT;
        printf( "%s: %s\n", SERVERNAME, NO_ROOM );
        RISK_SendMessage( iCommLink, &msg );
        _SRV_CleanupRegistration( );
        return;
        }

    /* Send its ID */
    msgClientIdent.iClientID = iNewClient;
    msgClientIdent.msgId = MSG_CLIENTIDENT;
    RISK_SendMessage( iCommLink, ( const Msgs * ) &msgClientIdent );
    printf( "%s: %s \"%s\" %s\n",
            SERVERNAME,
            NEW_CLIENT,
            strClientAddress,
            HAS_REGISTERED );

    /* Set up its fields */
    CLIENTS_SetCommLink( iNewClient, iCommLink );


    CLIENTS_SetAddress( iNewClient, strClientAddress );

    CLIENTS_SetType( iNewClient, iClientType );
    CLIENTS_SetSpecies( iNewClient, iClientSpecies );

    /* Remember to look for messages from it with select( ),
     * and update the maximum file descriptor.
     */

    FD_SET( iCommLink, &fdBackup );
    iMaxFileDescUsed = MAX( iMaxFileDescUsed, iCommLink );

    /* Let the new client know about old clients, if needed */
    /* should be passed client...*/
    CLIENTS_TellAboutOthers( iCommLink, iClientType );

    /* Let all other clients know about it */
    snprintf( buf,
              sizeof( buf ),
              "%s %s",
              iClientType == CLIENT_AI ? NEW_AI : NEW_CLIENT, HAS_REGISTERED );
    msgMess.strMessage = buf;
    msgMess.iFrom      = FROM_SERVER;
    msgMess.iTo        = DST_OTHER;
    msgMess.msgId      = MSG_MESSAGEPACKET;
    SRV_CompleteBroadcast( iNewClient, ( const Msgs * ) &msgMess );

    /* And tell the client that just joined about itself */
    snprintf( buf,
              sizeof( buf ),
              "%s \"%s\".",
              HELLO_NEW,
              CLIENTS_GetAddress( iNewClient ) );
    RISK_SendMessage( CLIENTS_GetCommLinkOfClient( iNewClient ),
                      ( const Msgs * ) &msgMess );

    /* Tell the new client to pop up its registration box, if we
     * are not in game playing mode.
     */

    if ( ( iGameState == GAME_REGISTERING )
         && ( iClientType != CLIENT_AI ) )
        {
        msg.msgId = MSG_POPUPREGISTERBOX;
        RISK_SendMessage( iCommLink, &msg );
        }


    /* If there are any players, let the new client
     * know about them, so that it can set up the
     * colors and other things.
     */

    if ( RISK_GetNumPlayers( )
         || ( RISK_GetNumSpecies( ) > 1 ) )
        {
        /* Let the clients know the situation */
        snprintf( buf, sizeof( buf ), "%s %s", CLIENT, BEING_UPDATED );
        SRV_BroadcastTextMessage( buf );

        if ( iGameState == GAME_REGISTERING )
            {
            RISK_ReplicateRegistrationData( iCommLink );
            }
        else if ( ( iGameState == GAME_PLAYING )
                  || ( iGameState == GAME_FORTIFYING ) )
            {
            MsgTurnNotify  msgTurnNotify;

            RISK_ReplicateAllData( iCommLink );

            /* Let the client know who's turn it is currently */
            msgTurnNotify.iPlayer = iTurn;
            msgTurnNotify.iClient =
                RISK_GetClientOfPlayer( msgTurnNotify.iPlayer );
            msgTurnNotify.msgId = MSG_TURNNOTIFY;
            RISK_SendMessage( iCommLink,
                              ( const Msgs * ) &msgTurnNotify );
            }
        }

    /* Don't need these anymore */
    if ( ( pDeleteMe == pMess ) && ( pDeleteMe != NULL ) )
        {
        NET_DeleteMessage( pDeleteMe );
        }
    else if ( pMess )
        {
        NET_DeleteMessage( pMess );
        }
    }


/**
 * Handles client deregistration.
 */
void SRV_HandleDEREGISTERCLIENT( Int32 iClient )
    {
    assert( ( iClient >= 0 )
            && ( iClient < MAX_CLIENTS  ) );

    /* We need to free the client */
    SRV_FreeClient( iClient );
    }


/**
 * Terminates the program.
 */
void UTIL_ExitProgram( Int32 iExitValue )
    {
    /* Don't let any more data be received on this socket */
    shutdown( iServerCommLink, 0 );

    exit( iExitValue );
    }


/**
 * Server signal handler.
 */
RETSIGTYPE SRV_HandleSignals( Int32 iParam )
    {
    Msgs    msg;
    Int32   i;

    /* printf( "Ouch, I've been killed...\n" ); */
    printf( "O! I am slain!\n" );

    if ( ( ( iGameState != GAME_FORTIFYING )
           && ( iGameState != GAME_PLAYING   ) )
         || ( ( iParam != SIGHUP )
              && ( iParam != SIGINT )
              && ( iParam != SIGQUIT )
              && ( iParam != SIGKILL ) )
         || ( CLIENTS_GetNumClients( ) <= 0 ) )
        {
        SRV_HandleEXIT( -1 );
        }

    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        if ( ( CLIENTS_GetAllocationState( i ) == ALLOC_COMPLETE )
             && ( CLIENTS_GetStartState( i ) == TRUE ) )
            {
            msg.msgId = MSG_DEREGISTERCLIENT;
            RISK_SendMessage( CLIENTS_GetCommLinkOfClient( i ),
                              &msg );
            close( CLIENTS_GetCommLinkOfClient( i ) );
            }
        }
    fRememberKilled = TRUE;
    select( iMaxFileDescUsed + 1,
            ( fd_set * ) &fdSet,
            ( fd_set * ) 0,
            ( fd_set * ) 0,
            NULL );
    signal( iParam, SRV_HandleSignals );
    }

/**
 * Totally removes the given client from the game.
 *
 * \b Notes:
 * \arg tony: to move to clients.c, fdbackup is problem
 */
void SRV_FreeClient( Int32 iClient )
    {
    Int32           i;
    Int32           iNumPlayers;
    MsgFreePlayer   msgFreePlayer;

    /* Get rid of the client.  We do this first so that broadcasts
     * won't include this client.  This would normally result in
     * a SIGPIPE, if the exiting client had already shut down its
     * sockets, but since we ignore SIGPIPE, it would result in
     * a failed broadcast, which is technically wrong.
     */

    CLIENTS_SetAllocationState( iClient, ALLOC_NONE );
    CLIENTS_SetNumClients( CLIENTS_GetNumClients( ) - 1 );

    /* Note that the client won't be participating in the game */
    CLIENTS_SetStartState( iClient, FALSE );

    /* Reset the fields, close the sockets */
    close( CLIENTS_GetCommLinkOfClient( iClient ) );

    /* Remove the socket from the fd_set */
    FD_CLR( CLIENTS_GetCommLinkOfClient( iClient ), &fdBackup );

    /* Free all of the players at the client that is being nixed.
     * The client would call CLNT_FreePlayer to do this, which
     * is an RPC type call that sends MSG_FREEPLAYER to the
     * server ( that's me! )  So what we do here is manufacture a
     * MSG_FREEPLAYER and call the handler for it.  A bit roundabout,
     * perhaps, but nevertheless a consistant way of dealing with
     * freeing players.  Could probably be improved.  Notice that
     * we get the Nth player, and _not_ the Nth live player.
     * Note also, that freeing a player has the side effect of
     * changing the value of RISK_GetNumPlayers( ).
     */

    iNumPlayers = RISK_GetNumPlayers( );

    for ( i = iNumPlayers - 1; i >= 0; i-- )
        {
        if ( RISK_GetClientOfPlayer( RISK_GetNthPlayer( i ) ) == iClient )
            {
            msgFreePlayer.iPlayer = RISK_GetNthPlayer( i );
            SRV_HandleFREEPLAYER( ( void * ) &msgFreePlayer );
            }
        }

    /* Additionally, if the client is an AI client, then we have to
     * remove the species from the list of species.
     */

    if ( CLIENTS_GetType( iClient ) == CLIENT_AI )
        {
        SRV_FreeSpecies( CLIENTS_GetSpecies( iClient ) );
        }

    if ( fRememberKilled &&
         ( CLIENTS_GetNumClientsStarted( ) <= 0 ) )
        {
        SRV_HandleEXIT( -1 );
        }

    if ( iGameState == GAME_REGISTERING )
        {
        /* Reset the game, and possible start it */
        SRV_ResetGame( );
        SRV_AttemptNewGame( );
        }
    }


/**
 * Log a client failure.
 */
void SRV_LogFailure( CString strReason, Int32 iCommLink )
    {
    const Int32 iClient = CLIENTS_GetCommLinkToClient( iCommLink );

    assert( ( iClient >= -1 )
            && ( iClient < MAX_CLIENTS ) );

    /* This stops any more bogus messages from coming in from the
     * dead client, which probably has nothing interesting to say
     * anyway...
     */

    FD_CLR( iCommLink, &fdBackup );

    /* If this is -1, then it is not a known client */
    if ( iClient == -1 )
        {
        /* This shouldn't happen much */
        printf( ERR_COMMFAILED );
        close( iCommLink );
        }
    else
        {
        /* Log the failure, to be dealt with later. */
        CLIENTS_SetFailure( iClient, strReason );
        }
    }


/**
 * Tries to recover all client failures.
 */
void SRV_RecoverFailures( void )
    {
    Int32   iClient;
    Int32   count;
    char    buf[ 256 ];

    /* An optimization */
    if ( CLIENTS_GetNumFailures( ) == 0 )
        {
        return;
        }

    /* Go through all of the failures and handle them */
    for ( iClient = 0; iClient != MAX_CLIENTS; iClient++ )
        {
        if ( CLIENTS_GetFailureCount( iClient ) > 0 )
            {
            MsgNetPopup         msgNetPopup;
            MsgMessagePacket    msgMessagePacket;

            /* It is in the list of clients, so we need to destroy it.
             * Make as if we received a MSG_DEREGISTERCLIENT.
             */

            SRV_HandleDEREGISTERCLIENT( iClient );

            /* Tell the clients what's happening ( N.B. this broadcast call
             * has to be made AFTER the CLIENTS_Free call, otherwise, the
             * broadcast will try to send a message to the client that
             * just failed, and we will enter a recursive situation. )
             */

            count = CLIENTS_GetFailureCount( iClient );
            /* Information */
            snprintf( buf,
                      sizeof( buf ),
                      "\"%s\" has failed ( %d failure%s ).",
                      CLIENTS_GetFailureReason( iClient ),
                      count,
                      ( count > 1 ) ? "s" : "" );

            printf( "%s: %s\n",
                    SERVERNAME,
                    CLIENTS_GetFailureReason( iClient ) );
            msgNetPopup.strMessage = buf;
            msgNetPopup.msgId = MSG_NETPOPUP;
            SRV_BroadcastMessage( ( const Msgs * ) &msgNetPopup );

            msgMessagePacket.strMessage = buf;
            msgMessagePacket.iFrom      = FROM_SERVER;
            msgMessagePacket.iTo        = DST_ALLPLAYERS;
            msgMessagePacket.msgId      = MSG_MESSAGEPACKET;
            SRV_BroadcastMessage( ( const Msgs * ) &msgMessagePacket );

            /* Pheww... Finished. */
            printf( TXT_HANDLED_FAILURE );
            CLIENTS_ResetFailure( iClient );
            }
        }

    /* Reset this */
    CLIENTS_SetNumFailures( 0 );
    }


/**
 * Determines the next player to receive the turn.
 */
Int32 SRV_IterateTurn( void )
    {
    /* It may be that the player who's turn it just was killed a few
     * players, and the number of live players has gone down.  The way
     * players are organized is as if they were stacked on one top of
     * the next, with the 0th live player at the bottom.  In other words,
     * if the 2nd live player gets killed, then the 3rd live player becomes
     * the 2nd live player, etc.  This is done so that one can loop from the
     * [0th -- NumLivePlayer( )th] live player.  However, when calculating
     * turns, this may screw us up.  If is the 2nd live player's turn, and
     * during this turn the 1st live player gets killed, then the next turn
     * belongs to the 2nd live player ( who used to be the 3rd live player ).
     * So to calculate the next turn, we iterate through the players, NOT
     * the live players.  We simply search for the next live player.  In
     * case this sounds obvious, it probably is, but the code use to read:
     * "iTurn = ( iTurn+1 ) % RISK_GetNumLivePlayers( )" which is obviously wrong.
     */

    do
        {
        /* Go to the next player, wrap-around if we're at the end */
        iTurn = ( iTurn + 1 ) % MAX_PLAYERS;
        }
    while ( ( RISK_GetStateOfPlayer( iTurn ) == FALSE )
            || ( RISK_GetAllocationStateOfPlayer( iTurn ) != ALLOC_COMPLETE ) );

    /* Return the player who's turn it is */
    return iTurn;
    }


/**
 * Starts the allocation of a new species.
 */
Int32 SRV_AllocSpecies( void )
    { 
    Int32 i;

    /* Simply loop through the species looking for an empty slot.
     * Since they are created on demand, if the list is full, when
     * we get to the end the Dist. Obj. will build us a new one.
     */

    for ( i = 0;
          RISK_GetAllocationStateOfSpecies( i ) != ALLOC_NONE;
          i++ )
        {
        ; /* TwiddleThumbs */
        }

    /* Mark the species as allocated.  Otherwise, if the requesting ai Client
     * hasn't done anything with the new species, and another ai Client
     * is asking for one, the first ai Client doesn't set the name to be
     * something non-NULL, then the server will return the SAME species
     * number -- race condition!
     */

    RISK_SetAllocationStateOfSpecies( i, ALLOC_INPROGRESS );

    return i;
    }


/**
 * Frees a species.
 */
void SRV_FreeSpecies( Int32 i )
    {
    /* Reset the allocation state of the species */
    RISK_SetAllocationStateOfSpecies( i, ALLOC_NONE );
    }


/**
 * Restarts the game.
 */
void SRV_ResetGame( void )
    {
    /* Things we have to do when a game is restarted:
     *   1. Set the game state to be GAME_REGISTERING
     *   2. Actually reset the dist. obj.
     *   3. Create a new deck of cards.
     */

    iGameState = GAME_REGISTERING;

    /* Reset the game if it has not been already reset */
    if ( !fGameReset )
        {
        RISK_ResetGame( );
        fGameReset = TRUE;

        /* Get a new deck, destroy the old one. */
        DECK_Destroy( pCardDeck );
        pCardDeck = DECK_Create( NUM_COUNTRIES + 2 );
        }
    }


/**
 * Sanity checks and attempting to start a new game.
 */
void SRV_AttemptNewGame( void )
    {
    Int32   n;
    Int32   iIndex;
    Flag    haveHumans = FALSE;

    /* See if we can start a game */
    if ( ( iGameState != GAME_REGISTERING )
         || ( CLIENTS_GetNumClientsStarted( ) != CLIENTS_GetNumClients( ) )
         || ( RISK_GetNumPlayers( ) < 2 ) )
        {
        return;
        }

    for ( n = 0; n < CLIENTS_GetNumClientsStarted( ); n++ )
        {
        if ( CLIENTS_GetSpecies( n ) == SPECIES_HUMAN )
            {
            haveHumans = TRUE;
            break;
            }
        }

    if ( !haveHumans )
        {
        return;
        }

    assert( RISK_GetNumLivePlayers( ) >= 2 );

    printf( "%s: %s\n", SERVERNAME, BEGINNING );
    /* We're going to need to reset it before playing again, so mark this. */
    fGameReset = FALSE;

    iGameState = GAME_FORTIFYING;

    SRV_SetInitialArmiesOfPlayers( );
    SRV_SetInitialMissionOfPlayers( );

    /* pacman sez:
     * Must choose a player to be "first" before choosing initial territories,
     * otherwise things won't go right. Example: with 5 players, the order of
     * things must be as follows:
     * 1. Every player gets 25 armies
     * 2. We go around choosing territories. When that's all done, the first
     *    two players have 9 each and the other three players have 8 each.
     * 3. Enter "fortify" mode - AND IT IS NOW THE 3RD PLAYER'S TURN. To pick a
     *    random player at this point is Wrong( TM ).
     * 4. When the fortify mode is done, we'll be back at the first player
     *    again. He gets to attack first.
     * This order might get disrupted during the fortify stage, if the clients
     * are using bpk's multiple-place patch. We must remember who was first to
     * undo this disruption and enter 4 with iTurn being what it should be,
     * Which is what it would have been if the clients had fortified
     * one-at-a-time, which is iFirstPlayer. */
    iIndex = rand( ) % RISK_GetNumLivePlayers( );
    iFirstPlayer = RISK_GetNthLivePlayer( iIndex );
    iTurn = iFirstPlayer;

    /* Dole out the countries */
    SRV_DistributeCountries( );

    SRV_NotifyClientsOfTurn( iTurn );

    /* Reset this for the next time */
    for ( n = 0; n != MAX_CLIENTS; n++ )
        {
        if ( CLIENTS_GetAllocationState( n ) == ALLOC_COMPLETE )
            {
            CLIENTS_SetStartState( n, FALSE );
            }
        }
    }

/* End of file */
